'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from Bio import SeqIO, SearchIO
from ete3 import Tree
from multiprocessing import sharedctypes
from property_manager import lazy_property, cached_property
from PySAIS import sais
import ctypes
import gzip
import logging 
import multiprocessing as mp
import numba
import numpy as np
import os
import tables

from ._utils import LOG
from .alphabets import Alphabet
from .hierarchy import get_lca_hog_off


@numba.njit
def get_transform(k, DIGITS_AA):
    t = np.zeros(k, dtype=np.uint64)
    for i in numba.prange(k):
        t[i] = len(DIGITS_AA) ** (k - (i + 1))
    return t


class Index(object):
    def __init__(
        self, db, k=None, reduced_alphabet=False, nthreads=1
    ):
        # load database object
        self.db = db

        if k is None:
            if '/Index' not in self.db.db:
                # default k
                self.k = 6
            else:
                self.k = self.db.db.root.Index._v_attrs['k']
        else:
            self.k = k

        # set index related features
        self.alphabet = Alphabet(n=(21 if not reduced_alphabet else 13))

        # performance features
        self.nthreads = nthreads

        # filter based on species attributes
        self.sp_filter = np.full((len(self.db._sp_tab),), False)

        # corresponding taxonomic filter (used in validation)
        self.tax_filter = np.full((len(self.db._tax_tab),), False)

    ### useful for validation
    def set_species(self, sp_ii):
        self.sp_filter[:] = True
        for i in sp_ii:
            self.sp_filter[i] = False

    def hide_taxa(self, taxa, nwk_fn):
        '''
        enables to hide species of some taxa e.g. for a Validation purpose
        '''
        for taxon in taxa:
            
            # get all species and descending taxa of a given taxon
            hidden_taxa, hidden_species = self.get_clade_specific_taxa_species(nwk_fn, taxon)
            
            # find species offsets in sp_tab and turn on filter for these
            hidden_species_offsets = np.searchsorted(self.db._sp_tab.col('ID'), np.array(list(hidden_species)))
            for sp_off in hidden_species_offsets:
                self.sp_filter[sp_off] = True
            
            # find taxon offsets in tax_tab and turn on filter for these
            hidden_taxa_offsets = np.searchsorted(self.db._tax_tab.col('ID'), np.array(list(hidden_taxa)))
            for tax_off in hidden_taxa_offsets:
                self.tax_filter[tax_off] = True

    @staticmethod
    def get_clade_specific_taxa_species(nwk_fn, root_taxon):
        '''
        gather all taxa and species specific to a given clade
        '''
        stree = Tree(nwk_fn, format=1, quoted_node_names=True)

        pruned_stree = [x for x in stree.traverse() if x.name == root_taxon][0]

        taxa = set()
        species = set()

        for tl in pruned_stree.traverse():
            taxon = tl.name.encode('ascii')
            taxa.add(taxon)
            if tl.is_leaf():
                species.add(taxon)

        return np.array(sorted(taxa)), species

    ### same as in database class; easy access to data ###
    def _get_node_if_exist(self, node):
        if node in self.db.db:
            return self.db.db.get_node(node)

    @property
    def _table_idx(self):
        return self._get_node_if_exist('/Index/TableIndex')

    @property
    def _table_buff(self):
        return self._get_node_if_exist('/Index/TableBuffer')

    @property
    def _fam_count(self):
        return self._get_node_if_exist('/Index/FamCount')

    @property
    def _hog_count(self):
        return self._get_node_if_exist('/Index/HOGcount')

    ### main function to build the index ###
    def build_kmer_table(self):

        #assert self.db.mode in {"w", "a"}, "Index must be opened in write mode."
        assert self.db.mode == 'w', "Index must be opened in write mode."
        assert ('/Index' not in self.db.db), 'Index has already been computed'

        # build suffix array with option to translate the sequence buffer first
        sa = self._build_suffixarray(
            self.alphabet.translate(self.db._seq_buff[:]), len(self.db._prot_tab)
        )

        # Set nthreads, note: this only works before numba called first time!
        numba.config.NUMBA_NUM_THREADS = self.nthreads

        self._build_kmer_table(sa)

    @staticmethod
    def _build_suffixarray(seqs, n):
        # Build suffix array
        sa = sais(seqs)
        sa[:n].sort()  # Sort delimiters by position
        return sa

    def _build_kmer_table(self, sa):
        #@numba.njit(parallel=True) --> was breaking for PANTHER database
        @numba.njit
        def _compute_mask_and_filter(
            sa, sa_mask, sa_filter, k, n, prot2spoff, prot2hogoff, sp_filter
        ):
            """
            1. compute a mapper between suffixes and HOGs
            2. simultaneously, compute suffix filter for species and suffixe < k
            """
            for i in numba.prange(n):
                # leverages the sorted protein delimiters at the beggining of the sa to get the suffix offsets by protein
                s = (sa[i - 1] if i > 0 else -1) + 1
                e = sa[i] + 1
                sa_mask[s:e] = prot2hogoff[i]
                if sp_filter[prot2spoff[i]]:
                    sa_filter[s:e] = True
                else:
                    sa_filter[(e - k) : e] = True

        @numba.njit
        def _same_kmer(seq_buff, sa, kmer, jj, k):
            kmer_jj = seq_buff[sa[jj] : (sa[jj] + k)].view(np.uint8)
            return np.all(np.equal(kmer, kmer_jj))

        @numba.njit
        def _compute_many_lca_hogs(hog_offsets, fam_offsets, hog_parents):
            """
            compute lca hogs for a list of hogs and their families
            """
            # as many lca hogs as unique families
            lca_hogs = np.zeros((np.unique(fam_offsets).size,), dtype=np.int64)

            # keep track of family, the corresponding hogs and the lca hog offset
            curr_fam = fam_offsets[0]
            curr_hogs = list(
                hog_offsets[0:1]
            )  # set the type of items in list to integers
            lca_off = 0

            for i in range(1, len(hog_offsets)):
                fam = fam_offsets[i]
                # wait to have all hogs of the family between computing the lca hog
                if fam == curr_fam:
                    curr_hogs.append(hog_offsets[i])
                else:
                    lca_hogs[lca_off] = get_lca_hog_off(curr_hogs, hog_parents)
                    curr_hogs = list(hog_offsets[i : i + 1])
                    curr_fam = fam
                    lca_off += 1

            # last family
            lca_hogs[lca_off] = get_lca_hog_off(curr_hogs, hog_parents)
            return lca_hogs

        @numba.njit
        def _compute_kmer_table(
            sa,
            seq_buff,
            sa_mask,
            hog_fams,
            hog_parents,
            table_idx,
            table_buff,
            fam_kmer_counts,
            hog_kmer_counts,
            k,
            DIGITS_AA,
            DIGITS_AA_LOOKUP,
        ):

            ii = 0  # pointer to sa offset of kk
            kk = 0  # pointer to k-mer (offset in table_idx)
            ii_table_buff = 0  # pointer to offset in table_buff (~rows of k-mer table)
            trans = get_transform(k, DIGITS_AA)
            while ii < len(sa):
                ## compute the new k-mer (kk1)
                kmer = seq_buff[sa[ii] : (sa[ii] + k)].view(np.uint8)
                kk1 = 0
                for i in range(k):
                    kk1 += int(DIGITS_AA_LOOKUP[kmer[i]] * trans[i])

                ## find offset of new k-mer in sa (jj)
                # first in windows of 100s
                jj = min(ii + 100, len(sa))
                while (jj < len(sa)) and _same_kmer(seq_buff, sa, kmer, jj, k):
                    jj = min(jj + 100, len(sa))

                # then, refine with binary search
                lo = max(ii, (jj - 100) + 1)
                hi = jj
                while lo < hi:
                    m = int(np.floor((lo + hi) / 2))
                    if _same_kmer(seq_buff, sa, kmer, m, k):
                        lo = m + 1
                    else:
                        hi = m
                jj = lo

                ## compute LCA HOGs containing current k-mer (kk)
                # get the hog offsets for each suffix containing kk
                hog_offsets = np.unique(sa_mask[ii:jj])

                # get the corresponding fam offsets
                fam_offsets = hog_fams[hog_offsets]

                # compute the LCA hog offsets
                lca_hog_offsets = _compute_many_lca_hogs(
                    hog_offsets, fam_offsets, hog_parents
                )

                # store kmer counts of fams and lca hogs
                fam_kmer_counts[np.unique(fam_offsets)] = (
                    fam_kmer_counts[np.unique(fam_offsets)] + 1
                )
                hog_kmer_counts[lca_hog_offsets] = hog_kmer_counts[lca_hog_offsets] + 1

                ## store LCA HOGs in table buffer
                nr_lca_hog_offsets = len(lca_hog_offsets)
                table_buff[
                    ii_table_buff : ii_table_buff + nr_lca_hog_offsets
                ] = lca_hog_offsets

                ## store buffer offset in table index at offset corresponding to k-mer integer encoding
                table_idx[kk : kk1 + 1] = ii_table_buff

                ## find buffer offset of new k-mer in table index
                ii_table_buff += nr_lca_hog_offsets
                ii = jj
                kk = kk1 + 1

            # fill until the end
            table_idx[kk:] = ii_table_buff
            return ii_table_buff

        LOG.debug(" - filter suffix array and compute its HOG mask")
        n = len(self.db._prot_tab)
        sa_mask = np.zeros(sa.shape, dtype=np.uint64)
        sa_filter = np.zeros(sa.shape, dtype=np.bool)

        _compute_mask_and_filter(
            sa,
            sa_mask,
            sa_filter,
            self.k,
            n,
            self.db._prot_tab.col("SpeOff"),
            self.db._prot_tab.col("HOGoff"),
            self.sp_filter,
        )

        # before filtering the sa, reorder and reverse the suffix filter
        sa = sa[~sa_filter[sa]]

        # filter and reorder the mask according to this filtered sa
        sa_mask = sa_mask[sa]

        LOG.debug(" - compute k-mer table")
        table_idx = np.zeros(
            (len(self.alphabet.DIGITS_AA) ** self.k + 1), dtype=np.uint64
        )

        # initiate buffer of size sa_mask, which is maximum size if all suffixes are from different HOGs
        table_buff = np.zeros((len(sa_mask)), dtype=np.uint64)

        fam_kmer_counts = np.zeros(len(self.db._fam_tab), dtype=np.uint64)
        hog_kmer_counts = np.zeros(len(self.db._hog_tab), dtype=np.uint64)

        ii_table_buff = _compute_kmer_table(
            sa,
            self.db._seq_buff[:],
            sa_mask,
            self.db._hog_tab.col("FamOff"),
            self.db._hog_tab.col("ParentOff"),
            table_idx,
            table_buff,
            fam_kmer_counts,
            hog_kmer_counts,
            self.k,
            self.alphabet.DIGITS_AA,
            self.alphabet.DIGITS_AA_LOOKUP,
        )

        # remove extra space
        table_buff = table_buff[:ii_table_buff]

        LOG.debug(" - write k-mer table")
        idx = self.db.db.create_group('/', 'Index', 'hog indexes')
        idx._f_setattr('k', self.k)
        self.db.db.create_carray(idx, "TableIndex", obj=table_idx, filters=self.db._compr)
        self.db.db.create_carray(
            idx, "TableBuffer", obj=table_buff, filters=self.db._compr
        )
        self.db.db.create_carray(
            idx, "FamCount", obj=fam_kmer_counts, filters=self.db._compr
        )  # for these, I can initialize them before...
        self.db.db.create_carray(
            idx, "HOGcount", obj=hog_kmer_counts, filters=self.db._compr
        )


class SequenceBuffer(object):
    """
    to load sequences from db or files
    adapted from alex
    """

    def __init__(self, seqs=None, ids=None, db=None, fasta_file=None):
        self.prot_off = 0
        if seqs is not None:
            self.add_seqs(*seqs)
            self.ids = np.array(ids) if ids else np.array(range(len(seqs)))
        elif db is not None:
            self._prot_tab = db._prot_tab
            self._seq_buff = db._seq_buff
            self.load_from_db()
        elif fasta_file is not None:
            seqs = self.parse_fasta(fasta_file)
            self.add_seqs(*seqs)
        else:
            raise ValueError(
                "need to pass array of seqs or db to load from to SequenceBuffer"
            )

    def __getstate__():
        return (self.prot_nr, self.n, self.buff_shr, self.buff_idx_shr)

    def __setstate__(state):
        (self.prot_nr, self.n, self.buff_shr, self.buff_idx_shr) = state

    def parse_fasta(self, fasta_file):
        ids = []
        seqs = []
        func = open if not fasta_file.endswith('.gz') else gzip.open
        with func(fasta_file, 'rt') as fp:
            for rec in SeqIO.parse(fp, "fasta"):
                ids.append(rec.id)
                seqs.append(str(rec.seq))
        self.ids = np.array(ids)
        return seqs

    def add_seqs(self, *seqs):
        self.prot_nr = len(seqs)
        self.n = self.prot_nr + sum(len(s) for s in seqs)

        self.buff_shr = sharedctypes.RawArray(ctypes.c_uint8, self.n)
        self.buff[:] = np.frombuffer(
            (" ".join(seqs) + " ").encode("ascii"), dtype=np.uint8
        )

        self.buff_idx_shr = sharedctypes.RawArray(ctypes.c_uint64, self.prot_nr + 1)
        for i in range(len(seqs)):
            self.idx[i + 1] = len(seqs[i]) + 1 + self.idx[i]

    def load_from_db(self):
        self.prot_nr = len(self._prot_tab)
        self.n = len(self._seq_buff)

        self.buff_shr = sharedctypes.RawArray(ctypes.c_uint8, self.n)
        self.buff[:] = self._seq_buff[:].view(np.uint8)

        self.buff_idx_shr = sharedctypes.RawArray(ctypes.c_uint64, self.prot_nr + 1)
        # self.idx[:-1] = self._prot_tab.cols.SeqOff[:]
        self.idx[:-1] = self._prot_tab[:]["SeqOff"]
        self.idx[-1] = self.n

        # store offsets as ids
        self.ids = np.arange(
            self.prot_off, self.prot_off + self.prot_nr, dtype=np.uint64
        )[:, None]

    @lazy_property
    def buff(self):
        return np.frombuffer(self.buff_shr, dtype=np.uint8).reshape(self.n)

    @lazy_property
    def idx(self):
        return np.frombuffer(self.buff_idx_shr, dtype=np.uint64).reshape(
            self.prot_nr + 1
        )

    # @lazy_property
    # def prot_offsets(self):
    #     return np.arange(self.prot_off, self.prot_off + self.prot_nr, dtype=np.uint64)

    def __getitem__(self, i):
        s = int(self.idx[i])
        e = int(self.idx[i + 1] - 1)
        return self.buff[s:e].tobytes().decode("ascii")


class QuerySequenceBuffer(SequenceBuffer):
    """
    get a sliced version of the sequence buffer object for a given species in db
    TO DO: put an __super__
    """

    def __init__(self, db, query_sp):
        self.query_sp = (
            query_sp if isinstance(query_sp, bytes) else query_sp.encode("ascii")
        )
        self.set_query_prot_tab(db)
        self.filter_query_seq_buff(db)
        self.load_from_db()

    def set_query_prot_tab(self, db):
        sp_off = np.searchsorted(db._sp_tab.col("ID"), self.query_sp)
        sp_ent = db._sp_tab[sp_off]
        self.prot_off = sp_ent["ProtOff"]
        self._prot_tab = db._prot_tab[self.prot_off : self.prot_off + sp_ent["ProtNum"]]

    def filter_query_seq_buff(self, db):
        self._seq_buff = db._seq_buff[
            self._prot_tab[0]["SeqOff"] : self._prot_tab[-1]["SeqOff"]
            + self._prot_tab[-1]["SeqLen"]
        ]
        # initialize sequence buffer offset
        self._prot_tab["SeqOff"] -= db._prot_tab[self.prot_off]["SeqOff"]



import sys
omamer_path = sys.argv[1]
sys.path.insert(0, omamer_path)

import os
import omamer
import numpy as np
from tqdm import tqdm

def run_validation_pipeline(
    working_path, root_taxon, min_hog_size, include_younger_fams, oma_db_fn, nwk_fn, k, reduced_alphabet, query_sp2hidden_taxa,
    query_sp, subfamily_query_sp2thresholds, focal_taxon, bin_num, val_mode, neg_root_taxon, chunksize):
    '''
    including automatic filenames
    '''
    out_path = '{}{}_{}families/'.format(working_path, root_taxon, 'only_' if not include_younger_fams else '')
    if not os.path.exists(out_path):
        os.mkdir(out_path)
    db_fn = '{}{}_{}wo_{}.h5'.format(out_path, root_taxon, 'younfams_' if include_younger_fams else '', '_'.join(['_'.join(x.split()) for x in query_sp2hidden_taxa[query_sp]]))
    va_fn = '{}_{}_{}.h5'.format(db_fn.split('.h5')[0], neg_root_taxon if neg_root_taxon else 'Random', '_'.join(query_sp.split()))

    va = _run_validation_pipeline(
        db_fn, root_taxon, min_hog_size, include_younger_fams, oma_db_fn, nwk_fn, k, reduced_alphabet, query_sp2hidden_taxa[query_sp], 
        va_fn, query_sp, subfamily_query_sp2thresholds[query_sp], root_taxon, bin_num, val_mode, neg_root_taxon, chunksize)
    
    return va

def _run_validation_pipeline(
    db_fn, root_taxon, min_hog_size, include_younger_fams, oma_db_fn, nwk_fn, k, reduced_alphabet, hidden_taxa,
    va_fn, query_sp, thresholds, focal_taxon, bin_num, val_mode, neg_root_taxon, chunksize):
    
    # build or reload database
    if not os.path.isfile(db_fn):
        db = omamer.DatabaseFromOMA(
            filename=db_fn, root_taxon=root_taxon, min_prot_nr=min_hog_size, 
            include_younger_fams=include_younger_fams, mode='w')
        
        # load sequences from OMA database
        db.build_database(oma_db_fn, nwk_fn)# build index
        
        # build k-mer table
        ki = omamer.Index(db, k=k, reduced_alphabet=reduced_alphabet)
        ki.hide_taxa(hidden_taxa, nwk_fn)
        ki.build_kmer_table()

        db.close()
    
    # reload database
    db = omamer.DatabaseFromOMA(
        filename=db_fn, root_taxon=root_taxon, min_prot_nr=min_hog_size, 
        include_younger_fams=include_younger_fams)
    
    qbuff = omamer.QuerySequenceBuffer(db, query_sp)

    # run or reload validation
    if not os.path.isfile(va_fn):
                      
        # setup search and validation
        ms = omamer.MergeSearch(ki=db.ki, nthreads=1)  # fails for nthreads > 1 BECAUSE numba version is <0.50
        va = omamer.Validation(
            db, filename=va_fn, thresholds=thresholds, nwk_fn=nwk_fn, query_sp=query_sp, 
            focal_taxon=focal_taxon, bin_num=bin_num, val_mode=val_mode, neg_root_taxon=neg_root_taxon, 
            max_query_nr=qbuff.prot_nr, neg_query_file='{}.fa'.format(va_fn.split('.')[0]), 
            oma_db_fn=oma_db_fn, nthreads=1)
        
        # search and validate
        ids = []
        seqs = []

        pbar = tqdm(desc='Searching')
        for i, q in enumerate(qbuff.ids):
            ids.append(q[0])
            seqs.append(qbuff[i])
            if len(ids) == chunksize:
                # search and validate the chunk
                ms.merge_search(seqs=seqs, ids=ids)
                va.validate(ms)     

                pbar.update(len(ids))
                ids = []
                seqs = []

        # search and validate last chunk
        ms.merge_search(seqs=seqs, ids=ids)
        va.validate(ms)
        if len(ids) > 0:
            pbar.update(len(ids))

        # close stuff
        pbar.close()
        va.va.close()
        
    va = omamer.Validation(
        db, filename=va_fn, thresholds=thresholds, nwk_fn=nwk_fn, query_sp=query_sp, 
        focal_taxon=focal_taxon, bin_num=bin_num, val_mode=val_mode, neg_root_taxon=neg_root_taxon, 
        max_query_nr=qbuff.prot_nr, neg_query_file='{}.fa'.format(va_fn.split('.')[0]), 
        oma_db_fn=oma_db_fn, nthreads=1)
    
    return va

working_path = sys.argv[2]
root_taxon = sys.argv[3]
min_hog_size = 6
include_younger_fams = True if (sys.argv[4] == 'True') else False
oma_path = sys.argv[5]
oma_db_fn = os.path.join(oma_path, "OmaServer.h5")
nwk_fn = os.path.join(oma_path, "speciestree.nwk")
k = 6
reduced_alphabet = False

query_sp = ' '.join(sys.argv[6].split('_'))

query_species = ['Ornithorhynchus anatinus', 'Lepisosteus oculatus', 'Branchiostoma floridae', 'Branchiostoma lanceolatum']
query_hidden_taxa = [['Ornithorhynchus anatinus'], ['Lepisosteus oculatus'], ['Branchiostoma'], ['Branchiostoma']]
query_sp2hidden_taxa = dict(zip(query_species, query_hidden_taxa))

omamer_thresholds = np.arange(0.001, 1.001, 0.005)
subfamily_query_sp2thresholds = {
    ('Ornithorhynchus anatinus'): omamer_thresholds,
    ('Lepisosteus oculatus'): omamer_thresholds,
    ('Branchiostoma floridae'): omamer_thresholds,
    ('Branchiostoma lanceolatum'): omamer_thresholds,
}

focal_taxon = "Metazoa"
bin_num = 1
val_mode = 'golike' 
neg_root_taxon_arg = sys.argv[7]
neg_root_taxon = None if (neg_root_taxon_arg == 'Random') else neg_root_taxon_arg

chunksize = int(sys.argv[8])

run_validation_pipeline(
    working_path, root_taxon, min_hog_size, include_younger_fams, oma_db_fn, nwk_fn, k, reduced_alphabet, query_sp2hidden_taxa,
    query_sp, subfamily_query_sp2thresholds, focal_taxon, bin_num, val_mode, neg_root_taxon, chunksize)



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
import numba
import numpy as np
import itertools
import tables


# generic functions for hierarchy
@numba.njit
def get_root_leaf_hog_offsets(off, parent_arr):
    """
    leverages parent pointers to gather parents until root
    """
    leaf_root = [off]
    parent = parent_arr[off]
    while parent != -1:
        leaf_root.append(parent)
        parent = parent_arr[parent]
    return np.array(leaf_root[::-1], dtype=np.uint64)


@numba.njit
def get_lca_hog_off(offsets, parent_arr):
    """
    compute the last common ancestor (lca) within a list of hog offsets from the same family
    """
    off_nr = len(offsets)

    # lca of one hog is itself
    if off_nr == 1:
        return np.uint64(offsets[0])

    else:
        # gather every root-to-leaf paths and keep track of the shortest one
        root_leaf_paths = []
        min_path = np.iinfo(np.int64).max
        for x in offsets:
            root_leaf = get_root_leaf_hog_offsets(x, parent_arr)
            root_leaf_paths.append(root_leaf)
            if len(root_leaf) < min_path:
                min_path = len(root_leaf)

        # if one hog is root, lca is root
        if min_path == 1:
            return np.uint64(root_leaf_paths[0][0])

        else:
            # hogs from root to leaves and stop at min_path
            mat = np.zeros((off_nr, min_path), dtype=np.int64)
            for i in range(off_nr):
                mat[i] = root_leaf_paths[i][:min_path]
            matT = mat.T

            # the lca hog is the one before hogs start diverging
            i = 0
            while i < min_path and np.unique(matT[i]).size == 1:
                i += 1
            return np.uint64(matT[i - 1][0])


def is_ancestor(hog1, hog2, hog2parent):
    """
    is hog1 ancestral to hog2
    """
    return hog1 in get_root_leaf_hog_offsets(hog2, hog2parent)


# Taxonomy


def children_tax(tax_off, tax_tab, ctax_buff):
    """
	collect direct children of a taxon
	"""
    tax_ent = tax_tab[tax_off]
    ctax_off = tax_ent["ChildrenOff"]
    return ctax_buff[ctax_off : ctax_off + tax_ent["ChildrenNum"]]


def get_lca_tax(tax_off, tax2parent, hidden_taxa):
    """
    get taxon from which tax_off has diverged
    """
    root_leaf = get_root_leaf_hog_offsets(tax_off, tax2parent)
    for x in root_leaf[::-1]:
        if x not in hidden_taxa:
            return x


def get_sister_taxa(tax_off, tax2parent, hidden_taxa, tax_tab, ctax_buff):

    lca_tax = get_lca_tax(tax_off, tax2parent, hidden_taxa)
    children = _children_tax(lca_tax, tax_tab, ctax_buff)

    return np.setdiff1d(children, tax_off)


def leaf_traverse(tax_off, tax_tab, ctax_buff, acc, leaf_fun):
    """
	extend to main traverse by adding postorder, preorder and midorder funs
	"""

    for ctax in children_tax(tax_off, tax_tab, ctax_buff):
        # reach species
        spe_off = tax_tab[ctax]["SpeOff"]
        if spe_off != -1:
            leaf_fun(spe_off, acc)
        else:
            leaf_traverse(ctax, tax_tab, ctax_buff, acc, leaf_fun)
    return acc


def get_descendant_species(tax_off, tax_tab, ctax_buff):
    def append_species(spe_off, list):
        list.append(spe_off)
        return list

    return np.array(
        sorted(leaf_traverse(tax_off, tax_tab, ctax_buff, [], append_species)),
        dtype=np.uint64,
    )


# HOGs


def _children_prot(hog_off, hog_tab, cprot_buff):
    """
    simply collect the proteins of a HOG
    """
    hog_ent = hog_tab[hog_off]
    cprot_off = hog_ent["ChildrenProtOff"]
    return cprot_buff[cprot_off : cprot_off + hog_ent["ChildrenProtNum"]]


def _children_hog(hog_off, hog_tab, chog_buff):
    """
	3 functions for children is silly
	"""
    hog_ent = hog_tab[hog_off]
    chog_off = hog_ent["ChildrenHOGoff"]
    return chog_buff[chog_off : chog_off + hog_ent["ChildrenHOGnum"]]


def traverse(hog_off, hog_tab, chog_buff, acc, leaf_fun, prefix_fun):

    prefix_fun(hog_off, acc)

    for chog in _children_hog(hog_off, hog_tab, chog_buff):

        # stop when no children
        if hog_tab[chog]["ChildrenHOGoff"] == -1:
            leaf_fun(chog, acc)
        else:
            traverse(chog, hog_tab, chog_buff, acc, leaf_fun, prefix_fun)

    return acc


def get_descendant_hogs(hog_off, hog_tab, chog_buff):
    def append_hog(hog_off, list):
        list.append(hog_off)
        return list

    descendant_hogs = traverse(hog_off, hog_tab, chog_buff, [], append_hog, append_hog)
    descendant_hogs.remove(hog_off)

    return np.array(np.unique(descendant_hogs), dtype=np.uint64)


def get_descendant_prots(descendant_hogs, hog_tab, cprot_buff):
    descendant_prots = []
    for x in descendant_hogs:
        descendant_prots.extend(list(_children_prot(x, hog_tab, cprot_buff)))
    return np.array(descendant_prots, dtype=np.uint64)


def get_descendant_species_taxoffs(
    hog_off, hog_tab, chog_buff, cprot_buff, prot2speoff, speoff2taxoff
):
    """
	traverse the HOG to collect all species 
	"""
    prot_offs = get_descendant_prots(
        np.append(get_descendant_hogs(hog_off, hog_tab, chog_buff), np.uint64(hog_off)),
        hog_tab,
        cprot_buff,
    )
    return speoff2taxoff[prot2speoff[prot_offs]]


# inparalog coverage new
def get_sispecies_candidates(tax_off, tax_tab, ctax_buff, hidden_taxa):

    # get each possible LCA speciation node
    tax_offs = _get_root_leaf_offsets(tax_off, tax_tab["ParentOff"])[::-1]

    sispecies_cands = []
    for i, to in enumerate(tax_offs[:-1]):

        parent = tax_offs[i + 1]

        # skip if parent taxa is hidden
        if parent in hidden_taxa:
            continue
        else:
            sistaxa = np.setdiff1d(_children_tax(parent, tax_tab, ctax_buff), to)
            sispecies = []
            for t in sistaxa:
                sispecies.append(list(get_descendant_species(t, tax_tab, ctax_buff)))
            sispecies_cands.append(list(itertools.chain(*sispecies)))

    return sispecies_cands


def get_sister_hogs(hog_off, hog_tab, chog_buff):
    parent = hog_tab["ParentOff"][hog_off]
    return np.setdiff1d(_children_hog(parent, hog_tab, chog_buff), hog_off)


def find_sisspecies(ortholog_species, inparalog_species, sispecies_cands):
    sispecies = []
    for ss_cand in sispecies_cands:
        if (
            len(
                set(list(ortholog_species) + list(inparalog_species)).intersection(
                    ss_cand
                )
            )
            > 0
        ):
            sispecies = ss_cand
            break
    return sispecies


def filter_proteins(proteins, protein_species, sisspecies):
    mask = np.array([x in sisspecies for x in protein_species])
    if mask.size > 0:
        return proteins[mask]
    else:
        return proteins


def calculate_inparalog_coverage(orthologs_f, inparalogs_f):
    n = inparalogs_f.size
    d = n + orthologs_f.size
    return n / d if n != 0 else 0


def compute_inparalog_coverage_new(
    qoff,
    query_ids,
    prot_tab,
    cprot_buff,
    hog_tab,
    chog_buff,
    hidden_taxa,
    sispecies_cands,
    verbose=0,
):

    poff = query_ids[qoff]
    hog_off = prot_tab[poff]["HOGoff"]

    leaf_root_hogs = _get_root_leaf_offsets(hog_off, hog_tab["ParentOff"])[::-1]

    for hog_off in leaf_root_hogs:
        hog_tax = hog_tab["TaxOff"][hog_off]
        hog_lcatax = hog_tab["LCAtaxOff"][hog_off]

        # duplication inside hidden taxa; move to parent
        if hog_tax in hidden_taxa:
            continue

        # duplication outside hidden taxa but not knowloedgeable because of all species members are hidden
        elif hog_lcatax in hidden_taxa:

            # sister HOGs at same taxon and with LCA taxon not hidden
            sis_hogs = get_sister_hogs(hog_off, hog_tab, chog_buff)
            sis_hogs = [
                h
                for h in sis_hogs
                if hog_tab["TaxOff"][h] == hog_tax
                and hog_tab["LCAtaxOff"][h] not in hidden_taxa
            ]

            # LCA node is a duplication (or several); consider all orthologs and inparalogs to calculate IC
            if sis_hogs:
                if verbose == 1:
                    print(hog_tab["ID"][sis_hogs])

                # collect proteins from sister HOGs sub-HOGs
                orthologs = get_descendant_prots(sis_hogs, hog_tab, cprot_buff)
                sis_inhogs = list(
                    itertools.chain(
                        *[
                            list(get_descendant_hogs(sh, hog_tab, chog_buff))
                            for sh in sis_hogs
                        ]
                    )
                )
                inparalogs = get_descendant_prots(sis_inhogs, hog_tab, cprot_buff)

                return calculate_inparalog_coverage(orthologs, inparalogs)

            # sister HOGs are not knowloedgeable
            else:
                continue

        # HOG is knowloedgeable
        else:
            if verbose == 1:
                print(hog_tab["ID"][hog_off])

            # collect orthologs and paralogs
            orthologs = _children_prot(hog_off, hog_tab, cprot_buff)
            inhogs = get_descendant_hogs(hog_off, hog_tab, chog_buff)
            inparalogs = get_descendant_prots(inhogs, hog_tab, cprot_buff)

            # get the species of these orthologs and paralogs
            ortholog_species = prot_tab["SpeOff"][orthologs]
            inparalog_species = prot_tab["SpeOff"][inparalogs]

            # find sister species
            sisspecies = find_sisspecies(
                ortholog_species, inparalog_species, sispecies_cands
            )

            # then filter out protein not descending from sister taxa
            orthologs_f = filter_proteins(orthologs, ortholog_species, sisspecies)
            inparalogs_f = filter_proteins(inparalogs, inparalog_species, sisspecies)

            return calculate_inparalog_coverage(orthologs_f, inparalogs_f)



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from property_manager import lazy_property, cached_property
import numba
import numpy as np
import tables
import os
import pandas as pd
import sys

from ._utils import LOG
from .index import get_transform, SequenceBuffer, QuerySequenceBuffer
from .hierarchy import get_descendant_hogs, get_descendant_prots

# --> will be renamed Search

# numba does not like nested methods 
@numba.njit
def get_fam_hog2parent(fam_ent, hog_tab):
    '''
    get HOG parent offsets of a single family
    '''
    hog_off = fam_ent['HOGoff']
    hog2parent_tmp = hog_tab['ParentOff'][hog_off:hog_off + fam_ent['HOGnum']]
    if hog2parent_tmp.size >1:
        return np.append(np.array([-1], dtype=np.int64), hog2parent_tmp[1:] - np.int64(hog_off))
    else:
        return hog2parent_tmp

@numba.njit
def get_fam_level_offsets(fam_ent, level_arr):
    '''
    get HOG level offsets of a single family
    '''
    level_off = fam_ent["LevelOff"]
    level_num = fam_ent["LevelNum"]
    fam_level_offsets = level_arr[level_off : np.int64(level_off + level_num + 2)]

    # because specific for a single family, reinitizialize offsets of family levels
    return fam_level_offsets - fam_level_offsets[0]

@numba.njit
def _max(x, y):
    return max(x, y)

@numba.njit
def _sum(x, y):
        return x + y 
    
@numba.njit
def cumulate_counts_1fam(
    hog_cum_counts, fam_level_offsets, hog2parent, cum_fun, prop_fun):

    current_best_child_count = np.zeros(hog_cum_counts.shape, dtype=np.uint16)

    # iterate over level offsets backward
    for i in range(fam_level_offsets.size - 2):
        x = fam_level_offsets[-i - 3 : -i - 1]

        # when reaching level, sum all hog counts with their best child count
        hog_cum_counts[x[0] : x[1]] = cum_fun(
            hog_cum_counts[x[0] : x[1]], current_best_child_count[x[0] : x[1]]
        )

        # update current_best_child_count of the parents of the current hogs
        for i in range(x[0], x[1]):
            parent_off = hog2parent[i]

            # only if parent exists
            if parent_off != -1:
                c = current_best_child_count[hog2parent[i]]
                current_best_child_count[hog2parent[i]] = prop_fun(
                    c, hog_cum_counts[i]
                )

@numba.njit
def store_bestpath(hog_offsets, parent_offsets, fam_bestpath, fam_hog_scores, pv_score):
    
    # keep HOGs descending from the best path
    cands = hog_offsets[fam_bestpath[parent_offsets]]

    # get the score of these candidates
    cands_scores = fam_hog_scores[cands]

    # find the candidate HOGs offsets with the higher count (>0) at this level
    if cands_scores.size > 0:

        # need smallest
        if pv_score:
            cands_offsets = np.where(cands_scores == np.min(cands_scores))[0]
        # need > 0 and max
        else:
            cands_offsets = np.where(
                (cands_scores > 0) & (cands_scores == np.max(cands_scores))
            )[0]

        # if a single candidate, update the best path. Else, stop because of tie
        if cands_offsets.size == 1:
            fam_bestpath[cands[cands_offsets]] = True
                
@numba.njit
def get_fam_bestpath(fam_hog_counts, fam_level_offsets, hog2parent, pv_score=False):
    
    fam_bestpath = np.full(fam_hog_counts.shape, False)

    # root-HOG
    hog_score = fam_hog_counts[0]
    if hog_score > 0:
        fam_bestpath[0] = True

    # loop through hog levels
    for i in range(1, fam_level_offsets.size - 2):
        x = fam_level_offsets[i : i + 2]
        hog_offsets = np.arange(x[0], x[1])

        # grab parents
        parent_offsets = hog2parent[hog_offsets]
        
        # store best path
        store_bestpath(hog_offsets, parent_offsets, fam_bestpath, fam_hog_counts, pv_score)
    
    return fam_bestpath

@numba.njit
def _norm_hog_correct_query_size(
    fam_hog_cumcounts, query_count, fam_level_offsets, hog2parent, fam_hog_counts, pv_score=False):

    fam_hog_scores = np.zeros(fam_hog_cumcounts.shape, dtype=np.float64)
    fam_bestpath = np.full(fam_hog_cumcounts.shape, False)
    revcum_counts = np.full(fam_hog_cumcounts.shape, query_count, dtype=np.uint16)

    # compute and store root-HOG score
    fam_hog_scores[0] = np.float64(fam_hog_cumcounts[0]) / query_count

    # set the root-HOG as best path
    fam_bestpath[0] = True
    
    # loop through hog levels
    for i in range(1, fam_level_offsets.size - 2):
        x = fam_level_offsets[i : i + 2]
        hog_offsets = np.arange(x[0], x[1])

        # grab parents
        parent_offsets = hog2parent[hog_offsets]

        # update query revcumcount, basically substracting parent counts from query count
        qh_count = revcum_counts[parent_offsets] - fam_hog_counts[parent_offsets]
        revcum_counts[hog_offsets] = qh_count

        # compute and store score
        fam_hog_scores[hog_offsets] = fam_hog_cumcounts[hog_offsets] / qh_count

        # store bestpath
        store_bestpath(hog_offsets, parent_offsets, fam_bestpath, fam_hog_scores, pv_score)

    return fam_hog_scores, fam_bestpath


class MergeSearch(object):
	def __init__(self, ki, nthreads=None, low_mem=False, include_extant_genes=False):
		assert ki.db.mode == "r", "Database must be opened in read mode."

		# load ki and db
		self.db = ki.db
		self.ki = ki

		# performance features
		self.nthreads = nthreads if nthreads is not None else os.cpu_count()
		self.low_mem = low_mem
		self.include_extant_genes = include_extant_genes
		self.query_sp = None

	@cached_property
	def trans(self):
	    return get_transform(self.ki.k, self.ki.alphabet.DIGITS_AA)

	@cached_property
	def table_idx(self):
	    x = self.ki._table_idx
	    return x[:] if not self.low_mem else x

	@cached_property
	def table_buff(self):
	    x = self.ki._table_buff
	    return x[:] if not self.low_mem else x

	@cached_property
	def fam_tab(self):
	    return self.db._fam_tab[:]

	@cached_property
	def hog_tab(self):
	    return self.db._hog_tab[:]

	@cached_property
	def level_arr(self):
	    return self.db._level_arr[:]	    

	@cached_property
	def max_hog_nr(self):
	    return np.int64(np.max(self.db._fam_tab.col('HOGnum')))

	def merge_search(self, seqs=None, ids=None, fasta_file=None):
		# load query sequences
		if seqs:
		    sbuff = SequenceBuffer(seqs=seqs, ids=ids)
		elif fasta_file:
		    sbuff = SequenceBuffer(fasta_file=fasta_file)

		queryFam_ranked, queryFam_scores, queryRankHog_bestpath, queryRankHog_scores = self._lookup(
		    sbuff.buff,
		    sbuff.idx,
		    self.trans,
		    self.table_idx,
		    self.table_buff,
		    self.ki.k,
		    self.ki.alphabet.DIGITS_AA_LOOKUP,
		    self.fam_tab,
		    self.hog_tab,
		    self.level_arr,
		    self.max_hog_nr,
		    top_n_fams = 1,
		    top_m_fams = 10,
		    cum_mode = 'max', 
		    score = 'correct_querysize')

		self._queryFam_ranked = queryFam_ranked
		self._queryFam_scores = queryFam_scores
		self._queryRankHog_bestpath = queryRankHog_bestpath
		self._queryRankHog_scores = queryRankHog_scores

		# store ids of sbuff
		self._query_ids = sbuff.ids.flatten()

	def output_results(self, threshold=0.1):

		def get_prot_ids(h):
		    desc_hogs = list(get_descendant_hogs(h, self.db._hog_tab, self.db._chog_arr)) + [h]
		    prot_ii = get_descendant_prots(desc_hogs, self.db._hog_tab, self.db._cprot_arr)
		    return self.db._prot_tab.read_coordinates(prot_ii, 'ID')

		qseqid = self._query_ids
		family = self._queryFam_ranked[:, 0]
		family_score = self._queryFam_scores[:, 0]

		# generate the dataframe
		def generate_results(threshold):
		    for i in np.argwhere(~np.isnan(family_score)).flatten():
		        # find best scoring subfamily
		        fam_hog_nr = self.db._fam_tab[family[i]]['HOGnum']
		        fam_bestpath = self._queryRankHog_bestpath[0, i, :fam_hog_nr]
		        fam_hog_scores = self._queryRankHog_scores[0, i, :fam_hog_nr]
		        best_j = None
		        best_s = np.inf
		        for j in np.argwhere(fam_bestpath).flatten():
		            s = fam_hog_scores[j]
		            if s < best_s and s >= threshold:
		                best_j = j
		                best_s = s
		        if best_j is not None:
		            hog_off = self.db._fam_tab[family[i]]['HOGoff']
		            # update best_j to get the subfamily offset
		            best_j = int(best_j + hog_off)
		            z = {'qseqid': qseqid[i],
		                 'family': self.db._hog_tab[self.db._fam_tab[family[i]]['HOGoff']]['OmaID'].decode('ascii'),
		                 'family-score': family_score[i],
		                 'subfamily': self.db._hog_tab[best_j]['OmaID'].decode('ascii'),
		                 'subfamily-score': best_s}
		            if self.include_extant_genes:
		                z['subfamily-geneset'] = ','.join(map(lambda x: x.decode('ascii'), get_prot_ids(best_j)))
		            yield z

		h = ['qseqid', 'family', 'family-score', 'subfamily', 'subfamily-score']
		if self.include_extant_genes:
		    h.append('subfamily-geneset')
		
		# if >0 family hit
		if np.max(family_score) >= threshold:
			return pd.DataFrame(generate_results(threshold))[h]
		else:
			return pd.DataFrame()

	@lazy_property
	def _lookup(self):
		def func(
		    seqs,
		    seqs_idx,
		    trans,
		    table_idx,
		    table_buff,
		    k,
		    DIGITS_AA_LOOKUP,
		    fam_tab,
		    hog_tab,
		    level_arr,
		    max_hog_nr,
		    top_n_fams = 1, 
		    top_m_fams = 1,
		    cum_mode = 'max',
		    score = 'correct_querysize'
		):   
		    '''
		    top_n_fams: number of family for which HOG scores are computed
		    top_m_fams: number of family for which family scores are computed before resorting
		    '''
		    # highest scoring families per query sorted by counts 
		    queryFam_ranked = np.zeros((len(seqs_idx) - 1, top_n_fams), dtype=np.uint32)
		    # corresponding k-mer counts
		    queryFam_scores = np.zeros((len(seqs_idx) - 1, top_n_fams), dtype=np.float64)
		    # scores and bestpath mask for HOGs of top_n_fam
		    queryRankHog_scores = np.zeros((top_n_fams, len(seqs_idx) - 1, max_hog_nr), dtype=np.float64)
		    queryRankHog_bestpath = np.zeros((top_n_fams, len(seqs_idx) - 1, max_hog_nr), dtype=np.bool8)    
		    # OPTION: store only HOGs on the bestpath

		    # to ignore k-mers with X (88 == b'X')
		    x_char = DIGITS_AA_LOOKUP[88]
		    # get a flag for k-mer with any X (= last k-mer + 1)
		    x_flag = table_idx.size - 1
		    # last_char = DIGITS_AA_LOOKUP[-1]
		    # x_flag = 0
		    # for j in range(k):
		    #     x_flag += trans[j] * last_char
		    # x_flag += 1

		    # iterate of sequences
		    for zz in numba.prange(len(seqs_idx) - 1):

		        ## get the query sequence
		        s = seqs[seqs_idx[zz] : np.int(seqs_idx[zz + 1] - 1)]
		        n_kmers = s.shape[0] - (k - 1)

		        # double check we don't have short peptides (of len < k)
		        # note: written in this order to provide loop-optimisation hint (?)
		        if n_kmers > 0:
		            pass
		        else:
		            continue

		        ## get the sequence unique k-mers
		        s_norm = DIGITS_AA_LOOKUP[s]
		        r = np.zeros(n_kmers, dtype=np.uint32)  # max kmer 7
		        for i in numba.prange(n_kmers):
		            # numba can't do np.dot with non-float
		            for j in numba.prange(k):
		                r[i] += trans[j] * s_norm[i + j]
		            # does k-mer contain any X?
		            x_seen = np.any(s_norm[i:i+k] == x_char)
		            # if yes, replace it by the x_flag
		            r[i] = r[i] if not x_seen else x_flag
		        r1 = np.unique(r)

		        # skip if one k-mer with X
		        if len(r1) > 1:
		            pass
		        elif r1[0] == x_flag:
		            continue

		        ## search k-mer table
		        hog_counts = np.zeros(hog_tab.size, dtype=np.uint16)
		        fam_counts = np.zeros(fam_tab.size, dtype=np.uint16)

		        # iterate unique k-mers
		        for m in numba.prange(r1.shape[0]):
		            kmer = r1[m]

		            # to ignore k-mers with X
		            if kmer == x_flag:
		                continue
		            else:
		                pass

		            # get mapping to HOGs
		            x = table_idx[kmer : kmer + 2]
		            hogs = table_buff[x[0] : x[1]]
		            fams = hog_tab['FamOff'][hogs]
		            hog_counts[hogs] += np.uint16(1)
		            fam_counts[fams] += np.uint16(1)

		        ### Get top n families
		        # get top m families from summed k-mer counts
		        top_fam = np.argsort(fam_counts)[::-1][:top_m_fams]
		        top_fam_counts = fam_counts[top_fam]

		        # cumulated queryHOG counts for the top m families
		        # small optimization: remember the fam_hog_cumcounts for the top n families for next step
		        if cum_mode == 'max':

		            # iterate over top n families
		            for fam_rank in numba.prange(top_m_fams):
		                fam_off = top_fam[fam_rank]
		                fam_ent = fam_tab[fam_off]
		                fam_hog_off = fam_ent['HOGoff']
		                fam_hog_nr = fam_ent['HOGnum']

		                # compute the cumulated HOG counts for that family
		                fam_hog_cumcounts = hog_counts[fam_hog_off:fam_hog_off + fam_hog_nr].copy()

		                fam_hog2parent = get_fam_hog2parent(fam_ent, hog_tab)
		                fam_level_offsets = get_fam_level_offsets(fam_ent, level_arr)

		                cumulate_counts_1fam(fam_hog_cumcounts, fam_level_offsets, fam_hog2parent, _sum, _max)

		                # replace summed counts by maxed counts (from highest scoring root-to-leaf path)
		                top_fam_counts[fam_rank] = fam_hog_cumcounts[0]

		        # compute family score (normalize by number of unique k-mers)
		        top_fam_scores = top_fam_counts / r1.size

		        # resort by score
		        idx = (-top_fam_scores).argsort()
		        top_fam = top_fam[idx]
		        top_fam_scores = top_fam_scores[idx]

		        # store them with corresponding scores
		        queryFam_ranked[zz, :top_n_fams] = top_fam[:top_n_fams]
		        queryFam_scores[zz, :top_n_fams] = top_fam_scores[:top_n_fams]

		        ### Compute HOG scores and bestpath for top n families
		        # iterate over top n families
		        for fam_rank in numba.prange(top_n_fams):
		            fam_off = top_fam[fam_rank]
		            fam_ent = fam_tab[fam_off]
		            fam_hog_off = np.int64(fam_ent['HOGoff'])  # slicing with int (vs. uint) enabled parallel=True
		            fam_hog_nr = np.int64(fam_ent['HOGnum'])

		            # compute the cumulated HOG counts for that family
		            fam_hog_counts = hog_counts[fam_hog_off:fam_hog_off + fam_hog_nr].copy()

		            fam_hog2parent = get_fam_hog2parent(fam_ent, hog_tab)
		            fam_level_offsets = get_fam_level_offsets(fam_ent, level_arr)

		            fam_hog_cumcounts = fam_hog_counts.copy()  # because I need both count and cum counts in correct query size

		            # don't know why I cannot one line this within numba
		            if cum_mode == 'max':
		                cumulate_counts_1fam(
		                    fam_hog_cumcounts, fam_level_offsets, fam_hog2parent, _sum, _max)
		            elif cum_mode == 'sum':
		                cumulate_counts_1fam(
		                    fam_hog_cumcounts, fam_level_offsets, fam_hog2parent, _sum, _sum)

		            if score == 'querysize':
		                # compute HOG score (normalize by number of unique k-mers)
		                fam_hog_scores = fam_hog_cumcounts / r1.size

		                # compute highest scoring root-to-leaf HOG path within the family
		                fam_bestpath = get_fam_bestpath(fam_hog_scores, fam_level_offsets, fam_hog2parent)

		            # problem of using multiple if elif here! but working above >?!?!?
		            #elif score == 'correct_querysize':
		            else:
		                fam_hog_scores, fam_bestpath = _norm_hog_correct_query_size(
		                    fam_hog_cumcounts, r1.size, fam_level_offsets, fam_hog2parent, fam_hog_counts)

		            queryRankHog_bestpath[fam_rank, zz, :fam_bestpath.size] = fam_bestpath
		            queryRankHog_scores[fam_rank, zz, :fam_hog_scores.size] = fam_hog_scores

		    return queryFam_ranked, queryFam_scores, queryRankHog_bestpath, queryRankHog_scores

		if not self.low_mem:
		    # Set nthreads, note: this only works before numba called first time!
		    numba.set_num_threads(self.nthreads)
		    return numba.jit(func, parallel=(True if self.nthreads > 1 else False), nopython=True, nogil=True)
		else:
		    return func



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from property_manager import lazy_property, cached_property
import numba
import numpy as np
import os
import sys
import tables

from ._utils import LOG
from .index import get_transform, SequenceBuffer, QuerySequenceBuffer


class FlatSearch(object):
    def __init__(self, ki, nthreads=None, low_mem=False):

        assert ki.db.mode == "r", "Database must be opened in read mode."

        # load ki and db
        self.db = ki.db
        self.ki = ki

        # performance features
        self.nthreads = nthreads if nthreads is not None else os.cpu_count()
        self.low_mem = low_mem

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.fs.close()

    # cached properties from Index
    @cached_property
    def trans(self):
        return get_transform(self.ki.k, self.ki.alphabet.DIGITS_AA)

    @cached_property
    def table_idx(self):
        x = self.ki._table_idx
        return x[:] if not self.low_mem else x

    @cached_property
    def table_buff(self):
        x = self.ki._table_buff
        return x[:] if not self.low_mem else x

    @cached_property
    def prot2hog(self):
        return self.db._prot_tab.col("HOGoff")

    @cached_property
    def hog2fam(self):
        return self.db._hog_tab.col("FamOff")
    
    # flat search function
    def flat_search(self, seqs=None, ids=None, fasta_file=None, query_sp=None):
        """
        point is to limit code duplication in FlatSearchValidation 
        """
        # load query sequences
        if seqs:
            sb = SequenceBuffer(seqs=seqs, ids=ids)
        elif fasta_file:
            sb = SequenceBuffer(fasta_file=fasta_file)
        elif query_sp:
            self.query_species = "".join(query_sp.split())
            sb = QuerySequenceBuffer(db=self.db, query_sp=self.query_species)

        return self._flat_search(sb)

    def _flat_search(self, sbuff):
        LOG.debug("look-up k-mer table")
        # fam_ranked, queryFam_count, queryFam_occur, queryHog_count, queryHog_occur, query_count, query_occur = self.lookup(sbuff)
        (
            fam_ranked,
            queryFam_count,
            queryFam_occur,
            queryHog_count,
            queryHog_occur,
            query_count,
            query_occur,
        ) = self._lookup(
            sbuff.buff,
            sbuff.idx,
            self.trans,
            self.table_idx,
            self.table_buff,
            self.db._hog_tab.col("FamOff"),
            self.db._fam_tab.nrows,
            self.db._hog_tab.nrows,
            self.ki.k,
            self.ki.alphabet.DIGITS_AA_LOOKUP,
        )

        self._fam_ranked = fam_ranked
        self._queryFam_count = queryFam_count
        self._queryFam_occur = queryFam_occur
        self._queryHog_count = queryHog_count
        self._queryHog_occur = queryHog_occur

        # store as vertical vectors
        self._query_count = query_count[:, None]
        self._query_occur = query_occur[:, None]

        # store ids of sbuff
        self._query_id = sbuff.ids.flatten()

    @lazy_property
    def _lookup(self):
        def func(
            seqs,
            seqs_idx,
            trans,
            table_idx,
            table_buff,
            hog2fam,
            n_fams,
            n_hogs,
            k,
            DIGITS_AA_LOOKUP,
        ):

            fam_results = np.zeros((len(seqs_idx) - 1, n_fams), dtype=np.uint32)
            fam_counts = np.zeros((len(seqs_idx) - 1, n_fams), dtype=np.uint16)
            fam_occur = np.zeros((len(seqs_idx) - 1, n_fams), dtype=np.uint32)  # ()
            hog_counts = np.zeros((len(seqs_idx) - 1, n_hogs), dtype=np.uint16)
            hog_occur = np.zeros((len(seqs_idx) - 1, n_hogs), dtype=np.uint32)  # ()
            query_counts = np.zeros((len(seqs_idx) - 1), dtype=np.uint16)  # ()
            query_occur = np.zeros((len(seqs_idx) - 1), dtype=np.uint32)  # ()

            x_char = DIGITS_AA_LOOKUP[88]  # 88 == b'X'
            x_kmer = 0
            for j in range(k):
                x_kmer += trans[j] * x_char
            x_kmer += 1

            for zz in numba.prange(len(seqs_idx) - 1):
                # grab query sequence
                s = seqs[seqs_idx[zz] : int(seqs_idx[zz + 1] - 1)]
                n_kmers = s.shape[0] - (k - 1)

                # double check we don't have short peptides
                # note: written in this order to provide loop-optimisation hint
                if n_kmers > 0:
                    pass
                else:
                    continue

                # parse into k-mers
                # TO DO: get locations simultaneously
                s_norm = DIGITS_AA_LOOKUP[s]

                r = np.zeros(n_kmers, dtype=np.uint32)  # max kmer 7
                for i in numba.prange(n_kmers):
                    # numba can't do np.dot with non-float
                    for j in range(k):
                        r[i] += trans[j] * s_norm[i + j]

                    x_seen = np.any(s_norm[i:i+k] == x_char)
                    r[i] = r[i] if not x_seen else x_kmer
                r1 = np.unique(r)

                if len(r1) > 1:
                    pass
                elif r1[0] == x_kmer:
                    continue

                # search hogs and fams
                hog_res = np.zeros(n_hogs, dtype=np.uint16)
                fam_res = np.zeros(n_fams, dtype=np.uint16)
                hog_occ = np.zeros(n_hogs, dtype=np.uint32)
                fam_occ = np.zeros(n_fams, dtype=np.uint32)
                query_occ = 0
                for m in numba.prange(r1.shape[0]):
                    kmer = r1[m]
                    if kmer < x_kmer:
                        pass
                    else:
                        continue
                    x = table_idx[kmer : kmer + 2]
                    hogs = table_buff[x[0] : x[1]]
                    fams = hog2fam[hogs]
                    hog_res[hogs] += np.uint16(1)
                    fam_res[fams] += np.uint16(1)
                    # kmer_occ is the nr of hogs/fams with the given kmer, used to compute its frequency
                    kmer_occ = (
                        hogs.size
                    )  # the built-in function len made a type error only when parallel was ON
                    # store for the set of query k-mers
                    query_occ += kmer_occ
                    # and for the set of intersecting k-mers between the query and every hogs and fams
                    hog_occ[hogs] += np.uint32(kmer_occ)
                    fam_occ[fams] += np.uint32(
                        kmer_occ
                    )  # I think here it should be len(fams) instead of len(hogs) ... # actually fine because one hog for one fam...

                # report results for families sorted by k-mer count
                t = np.argsort(fam_res)[::-1]
                fam_results[zz, :n_fams] = t
                # not sorting anymore the corresponding fam counts and occur
                fam_counts[zz, :n_fams] = fam_res
                fam_occur[zz, :n_fams] = fam_occ
                # fam_counts[zz, :n_fams] = fam_res[t]
                # fam_occur[zz, :n_fams] = fam_occ[t]
                # report raw results for hogs
                hog_counts[zz, :n_hogs] = hog_res
                hog_occur[zz, :n_hogs] = hog_occ
                # report results for the query
                query_counts[zz] = len(r1)
                query_occur[zz] = query_occ

            return (
                fam_results,
                fam_counts,
                fam_occur,
                hog_counts,
                hog_occur,
                query_counts,
                query_occur,
            )

        if not self.low_mem:
            # Set nthreads, note: this only works before numba called first time!
            numba.config.NUMBA_NUM_THREADS = self.nthreads
            return numba.jit(func, parallel=True, nopython=True, nogil=True)
        else:
            return func



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from Bio import SeqIO
from ete3 import Tree
from itertools import repeat
from property_manager import lazy_property
from tqdm import tqdm
import collections
import logging
import numpy as np
import os
import tables

from ._utils import LOG, is_progress_disabled

from .hierarchy import (
    get_root_leaf_hog_offsets,
    get_lca_hog_off,
    get_descendant_species_taxoffs,
)
from .index import Index


class Database(object):
    # define the format of the tables
    class ProteinTableFormat(tables.IsDescription):
        ID = tables.StringCol(255, pos=1, dflt=b"")
        SeqOff = tables.UInt64Col(pos=2)
        SeqLen = tables.UInt64Col(pos=3)
        SpeOff = tables.UInt64Col(pos=4)
        HOGoff = tables.UInt64Col(pos=5)
        FamOff = tables.UInt64Col(pos=6)

    class HOGtableFormat(tables.IsDescription):
        ID = tables.StringCol(255, pos=1, dflt=b"")
        FamOff = tables.UInt64Col(pos=2)
        TaxOff = tables.UInt64Col(pos=3)
        ParentOff = tables.Int64Col(pos=4)
        ChildrenHOGoff = tables.Int64Col(pos=5)
        ChildrenHOGnum = tables.Int64Col(pos=6)
        ChildrenProtOff = tables.Int64Col(pos=7)
        ChildrenProtNum = tables.Int64Col(pos=8)
        OmaID = tables.StringCol(255, pos=9, dflt=b"")
        LCAtaxOff = tables.UInt64Col(pos=10)

    class FamilyTableFormat(tables.IsDescription):
        ID = tables.UInt64Col(pos=1)
        TaxOff = tables.UInt64Col(pos=2)
        HOGoff = tables.UInt64Col(pos=3)
        HOGnum = tables.UInt64Col(pos=4)
        LevelOff = tables.UInt64Col(pos=5)
        LevelNum = tables.UInt64Col(pos=6)

    class SpeciesTableFormat(tables.IsDescription):
        ID = tables.StringCol(255, pos=1, dflt=b"")
        ProtOff = tables.UInt64Col(pos=2)
        ProtNum = tables.UInt64Col(pos=3)
        TaxOff = tables.UInt64Col(pos=4)

    class TaxonomyTableFormat(tables.IsDescription):
        ID = tables.StringCol(255, pos=1, dflt=b"")
        ParentOff = tables.Int64Col(pos=2)
        ChildrenOff = tables.Int64Col(pos=3)
        ChildrenNum = tables.Int64Col(pos=4)
        SpeOff = tables.Int64Col(pos=5)
        Level = tables.Int64Col(pos=6)

    def __init__(self, filename, root_taxon=None, mode='r', nthreads=1):
        assert (mode != 'w' or root_taxon), "A root_taxon must be defined when building the database"
        self.root_taxon = root_taxon
        self.filename = filename
        self.mode = mode

        self.nthreads = nthreads
        self._compr = tables.Filters(complevel=6, complib="blosc", fletcher32=True)

        if os.path.isfile(self.filename) and self.mode in {'a', 'w'}:
            LOG.warning('Overwriting database file ({})'.format(self.filename))

        self.db = tables.open_file(self.filename, self.mode, filters=self._compr)

        if '/Index' in self.db:
            self.ki = Index(self, nthreads=self.nthreads)

    def close(self):
        self.db.close()

    def __enter__(self):
        return self

    def __exit__(self, *_):
        self.close()

    def clean(self):
        '''
        close and remove hdf5 file
        '''
        self.__exit__()
        try:
            os.remove(self.filename)
        except FileNotFoundError:
            print("{} already cleaned".format(self.filename))

    ### method attributes to facilitate access to data stored in hdf5 file ###
    @property
    def _prot_tab(self):
        if "/Protein" in self.db:
            return self.db.root.Protein
        else:
            return self.db.create_table(
                "/", "Protein", self.ProteinTableFormat, filters=self._compr
            )

    @property
    def _hog_tab(self):
        if "/HOG" in self.db:
            return self.db.root.HOG
        else:
            return self.db.create_table(
                "/", "HOG", self.HOGtableFormat, filters=self._compr
            )

    @property
    def _fam_tab(self):
        if "/Family" in self.db:
            return self.db.root.Family
        else:
            return self.db.create_table(
                "/", "Family", self.FamilyTableFormat, filters=self._compr
            )

    @property
    def _sp_tab(self):
        if "/Species" in self.db:
            return self.db.root.Species
        else:
            return self.db.create_table(
                "/", "Species", self.SpeciesTableFormat, filters=self._compr
            )

    @property
    def _tax_tab(self):
        if "/Taxonomy" in self.db:
            return self.db.root.Taxonomy
        else:
            return self.db.create_table(
                "/", "Taxonomy", self.TaxonomyTableFormat, filters=self._compr
            )

    # arrays
    @property
    def _seq_buff(self):
        if "/SequenceBuffer" in self.db:
            return self.db.root.SequenceBuffer
        else:
            # initiate
            return self.db.create_earray(
                "/", "SequenceBuffer", tables.StringAtom(1), (0,), filters=self._compr
            )

    # The below arrays are of fixed length and created later as carray
    @property
    def _chog_arr(self):
        if "/ChildrenHOG" in self.db:
            return self.db.root.ChildrenHOG
        else:
            return None

    @property
    def _cprot_arr(self):
        if "/ChildrenProt" in self.db:
            return self.db.root.ChildrenProt
        else:
            return None

    @property
    def _ctax_arr(self):
        if "/ChildrenTax" in self.db:
            return self.db.root.ChildrenTax
        else:
            return None

    @property
    def _level_arr(self):
        if "/LevelOffsets" in self.db:
            return self.db.root.LevelOffsets
        else:
            return None

    ### functions common to DatabaseFromOMA and DatabaseFromFasta ###
    def initiate_tax_tab(self, stree_path):
        """
	except SpeOff column
        """
        assert self.mode in {"w", "a"}, "Database must be opened in write mode."

        def _parse_stree(stree_path, roottax):

            stree = Tree(stree_path, format=1, quoted_node_names=True)
            pruned_stree = [x for x in stree.traverse() if x.name == roottax][0]

            tax2parent = {}
            tax2children = {}
            tax2level = {}
            species = set()

            for tl in pruned_stree.traverse():
                tax = tl.name.encode('ascii')
                tax2parent[tax] = tl.up.name.encode('ascii') if tl.up else 'root'.encode('ascii')
                tax2children[tax] = [x.name.encode('ascii') for x in tl.children]
                tax2level[tax] = tl.get_distance(stree)
                if tl.is_leaf():
                    species.add(tl.name.encode('ascii'))

            return tax2parent, tax2children, tax2level, species

        # parse species tree
        tax2parent, tax2children, tax2level, species = _parse_stree(
            stree_path, self.root_taxon
        )

        # sort taxa and create mapper to their offsets
        sorted_taxa = sorted(tax2parent.keys())
        tax2taxoff = dict(zip(sorted_taxa, range(len(tax2parent))))

        # collect rows of taxonomy table and children buffer
        tax_rows = []
        children_buffer = []

        children_buffer_off = 0
        for tax in sorted_taxa:

            # must not be root
            par_off = (
                np.searchsorted(sorted_taxa, tax2parent[tax])
                if tax != self.root_taxon.encode("ascii")
                else -1
            )

            child_offsets = [np.searchsorted(sorted_taxa, x) for x in tax2children[tax]]

            # if no children, -1
            tax_rows.append(
                (
                    tax,
                    par_off,
                    children_buffer_off if child_offsets else -1,
                    len(child_offsets),
                    -1,
                    tax2level[tax],
                )
            )

            children_buffer.extend(child_offsets)
            children_buffer_off += len(child_offsets)

        # fill tax table
        self._tax_tab.append(tax_rows)
        self._tax_tab.flush()

        # store children taxa
        if not self._ctax_arr:
            self.db.create_carray(
                "/",
                "ChildrenTax",
                obj=np.array(children_buffer, dtype=np.int64),
                filters=self._compr,
            )

        return tax2taxoff, species

    def update_hog_and_fam_tabs(
        self, fam2hogs, hog2taxoff, hog2protoffs, hog2oma_hog=None
    ):
        assert self.mode in {"w", "a"}, "Database must be opened in write mode."

        def _get_parents(hog_off, hog_num, hogs):

            parents = []
            parent2child_hogs = collections.defaultdict(list)

            # sort alphabetically and track their true offsets
            alpha_hogs, alpha_hogs_offs = map(
                np.array, zip(*sorted(zip(hogs, range(hog_num))))
            )

            for i, hog_id in enumerate(hogs):
                child_hog = hog_off + i

                # use alphabetically sorted hogs here because np.searchsorted
                parent_hog = self.find_parent_hog(hog_id, alpha_hogs)
                if parent_hog != -1:
                    # map back to original hog offset and add the global hog_off
                    parent_hog = hog_off + alpha_hogs_offs[parent_hog]
                parents.append(parent_hog)

                if parent_hog != -1:
                    parent2child_hogs[parent_hog].append(child_hog)

            return parents, parent2child_hogs

        def _get_child_hogs(hog_off, hog_num, parent2child_hogs, child_hogs_off):

            children_offsets = []
            children_numbers = []
            children = []
            children_off = child_hogs_off

            for tmp_hog_off in range(hog_off, hog_off + hog_num):

                # get child hogs from the current HOG's offset
                tmp_children = parent2child_hogs.get(tmp_hog_off, [])
                children_num = len(tmp_children)

                # if >0 child hogs, keep track of offset, number and child hogs
                if children_num > 0:
                    children_offsets.append(children_off)
                    children.extend(tmp_children)

                # else, store -1 as offset and 0 as count
                else:
                    children_offsets.append(-1)

                # keep track of number of children
                children_numbers.append(children_num)

                # update children offset
                children_off += children_num

            return children_offsets, children_numbers, children, children_off

        def _get_child_prots(hogs, hog2protoffs, child_prots_off):

            children_offsets = []
            children_numbers = []
            children = []
            children_off = child_prots_off

            for hog in hogs:

                # get child prots from the current HOG id
                tmp_children = hog2protoffs.get(hog, [])
                children_num = len(tmp_children)

                # if >0 child prots, keep track of offset, number and child hogs
                if children_num > 0:
                    children_offsets.append(children_off)
                    children.extend(tmp_children)

                # else, store -1 as offset and 0 as count
                else:
                    children_offsets.append(-1)

                # keep track of number of children
                children_numbers.append(children_num)

                # update children offset
                children_off += children_num

            return children_offsets, children_numbers, children, children_off

        # main
        hog_rows = []
        fam_rows = []

        hog_off = 0
        fam_off = 0

        # initiate HOG and protein children arrays
        child_hogs = []
        child_prots = []
        child_hogs_off = 0
        child_prots_off = 0

        # initiate level array
        level_offsets = [0]
        level_offsets_off = 0

        for fam_id, hogs in sorted(fam2hogs.items(), key=lambda x: int(x[0])):

            ### hogs
            # sort by level
            hogs = sorted(hogs, key=lambda x: len(x.split(b".")))
            hog_num = len(hogs)
            hog_taxoffs = list(map(lambda x: hog2taxoff[x], hogs))

            # parents
            parents, parent2child_hogs = _get_parents(hog_off, hog_num, hogs)

            # levels
            hog_levels = list(map(lambda x: len(x.split(b".")), hogs))
            hog_level_offsets = (
                np.cumsum(np.unique(hog_levels, return_counts=True)[1]) + hog_off
            )
            hog_level_offsets_num = len(hog_level_offsets)
            level_offsets.extend(hog_level_offsets)

            # children hogs
            (
                child_hogs_offsets,
                child_hogs_numbers,
                tmp_child_hogs,
                child_hogs_off,
            ) = _get_child_hogs(hog_off, hog_num, parent2child_hogs, child_hogs_off)

            child_hogs.extend(tmp_child_hogs)

            # children prots
            (
                child_prots_offsets,
                child_prots_numbers,
                tmp_child_prots,
                child_prots_off,
            ) = _get_child_prots(hogs, hog2protoffs, child_prots_off)

            child_prots.extend(tmp_child_prots)

            # OMA hogs
            oma_hogs = (
                list(map(lambda x: hog2oma_hog[x], hogs))
                if hog2oma_hog
                else repeat(b"", hog_num)
            )

            hog_rows.extend(
                list(
                    zip(
                        hogs,
                        repeat(fam_off),
                        hog_taxoffs,
                        parents,
                        child_hogs_offsets,
                        child_hogs_numbers,
                        child_prots_offsets,
                        child_prots_numbers,
                        oma_hogs,
                        repeat(0),
                    )
                )
            )

            ### fams
            fam_rows.append(
                (
                    fam_id,
                    hog_taxoffs[0],
                    hog_off,
                    hog_num,
                    level_offsets_off,
                    hog_level_offsets_num,
                )
            )

            hog_off += hog_num
            fam_off += 1
            level_offsets_off += hog_level_offsets_num

        # fill family and HOG tables
        self._fam_tab.append(fam_rows)
        self._fam_tab.flush()
        self._hog_tab.append(hog_rows)
        self._hog_tab.flush()

        # store children simulaneously without previous initiation because I did not know the size of these
        if not self._chog_arr:
            self.db.create_carray(
                "/",
                "ChildrenHOG",
                obj=np.array(child_hogs, dtype=np.uint64),
                filters=self._compr,
            )
        if not self._cprot_arr:
            self.db.create_carray(
                "/",
                "ChildrenProt",
                obj=np.array(child_prots, dtype=np.uint64),
                filters=self._compr,
            )

        # store offsets of levels of hogs inside families. copy the last one to go reverse
        level_offsets.append(level_offsets[-1])
        if not self._level_arr:
            self.db.create_carray(
                "/",
                "LevelOffsets",
                obj=np.array(level_offsets, dtype=np.int64),
                filters=self._compr,
            )

    def add_speoff_col(self):
        """
	to the taxonomy table
	"""
        # load species and taxa
        species = self._sp_tab.col("ID")
        taxa = self._tax_tab.col("ID")

        # potential idx of each taxon in species
        species_idx = np.searchsorted(species, taxa)

        # add extra elem because np.searchsorted gave idx 1 too high
        species = np.append(species, b"")

        # mask for taxa being species
        species_mask = species[species_idx] == taxa

        # species offsets
        spe_offsets = np.arange(0, species.size - 1)

        # create SpeOff column
        speoff_col = np.full(taxa.size, -1)
        speoff_col[species_mask] = spe_offsets

        # update tax table
        self._tax_tab.modify_column(colname="SpeOff", column=speoff_col)

    def add_taxoff_col(self):
        """
		to the species table
		"""
        # load species and taxa
        species = self._sp_tab.col("ID")
        taxa = self._tax_tab.col("ID")

        tax_offsets = np.searchsorted(taxa, species)

        self._sp_tab.modify_column(colname="TaxOff", column=tax_offsets)

    def update_prot_tab(self, hog2protoffs):
        """
        add HOGoff and FamOff rows in the protein table
	"""
        assert self.mode in {"w", "a"}, "Database must be opened in write mode."

        def _get_protoffs2hogoff(hog2protoffs, hog2hogoff):
            protoff2hogoff = dict()
            for hog, protoffs in hog2protoffs.items():
                protoff2hogoff.update(dict(zip(protoffs, repeat(hog2hogoff[hog]))))
            return protoff2hogoff

        hog2hogoff = dict(zip(self._hog_tab.col("ID"), range(len(self._hog_tab))))
        protoffs2hogoff = _get_protoffs2hogoff(hog2protoffs, hog2hogoff)

        # get the hogoff column from it and the famoff col
        hogoff_col = list(zip(*sorted(protoffs2hogoff.items())))[1]
        famoff_col = self._hog_tab[np.array(hogoff_col)]["FamOff"]

        # replace the empty columns
        self._prot_tab.modify_column(colname="HOGoff", column=hogoff_col)
        self._prot_tab.modify_column(colname="FamOff", column=famoff_col)

    def add_lcataxoff_col(self):
        """
        compute the LCA taxon between the HOG species members
        """
        hog_tab = self._hog_tab[:]
        chog_buff = self._chog_arr[:]
        cprot_buff = self._cprot_arr[:]
        prot2speoff = self._prot_tab.col("SpeOff")
        speoff2taxoff = self._sp_tab.col("TaxOff")
        parent_arr = self._tax_tab.col("ParentOff")

        lcatax_offsets = np.zeros(hog_tab.size, dtype=np.uint64)

        for hog_off in range(hog_tab.size):
            species_tax_offs = get_descendant_species_taxoffs(
                hog_off, hog_tab, chog_buff, cprot_buff, prot2speoff, speoff2taxoff
            )
            lcatax_offsets[hog_off] = get_lca_hog_off(species_tax_offs, parent_arr)

        self._hog_tab.modify_column(colname="LCAtaxOff", column=lcatax_offsets)

    ### generic functions ###
    @staticmethod
    def parse_hogs(hog_id):
        if isinstance(hog_id, str):
            split_hog_id = hog_id.split(".")
            return [".".join(split_hog_id[: i + 1]) for i in range(len(split_hog_id))]
        elif isinstance(hog_id, bytes):
            split_hog_id = hog_id.split(b".")
            return [b".".join(split_hog_id[: i + 1]) for i in range(len(split_hog_id))]

    @staticmethod
    def find_parent_hog(hog_id, hogs):
        if isinstance(hog_id, str):
            split_hog = hog_id.split(".")
            parent_hog = -1
            if len(split_hog) > 1:
                parent_hog = np.searchsorted(hogs, ".".join(split_hog[:-1]))
            return parent_hog
        elif isinstance(hog_id, bytes):
            split_hog = hog_id.split(b".")
            parent_hog = -1
            if len(split_hog) > 1:
                parent_hog = np.searchsorted(hogs, b".".join(split_hog[:-1]))
            return parent_hog


class DatabaseFromOMA(Database):
    """
    Used to parse the OMA browser database file
    """
    def __init__(self, filename, root_taxon, min_prot_nr=6, include_younger_fams=True, mode='r'):
        super().__init__(filename, root_taxon, mode=mode)

        self.min_prot_nr = min_prot_nr
        self.include_younger_fams = include_younger_fams

    ### main function ###
    def build_database(self, oma_h5_path, stree_path):
        assert self.mode in {"w", "a"}, "Database must be opened in write mode."

        # build taxonomy table except the SpeOff column
        LOG.debug("initiate taxonomy table")
        tax2taxoff, species = self.initiate_tax_tab(stree_path)

        # load OMA h5 file
        h5file = tables.open_file(oma_h5_path, mode="r")

        LOG.debug("select and strip OMA HOGs")
        fam2hogs, hog2oma_hog, hog2tax = self.select_and_strip_OMA_HOGs(h5file)

        LOG.debug("fill sequence buffer, species table and initiate protein table")
        (
            fam2hogs,
            hog2protoffs,
            hog2tax,
            hog2oma_hog,
        ) = self.select_and_filter_OMA_proteins(
            h5file,
            fam2hogs,
            hog2oma_hog,
            hog2tax,
            species,
            self.min_prot_nr
        )

        LOG.debug(
            "add SpeOff and TaxOff columns in taxonomy and species tables, respectively"
        )
        self.add_speoff_col()
        self.add_taxoff_col()

        # mapper HOG to taxon offset
        hog2taxoff = {h: tax2taxoff.get(t, -1) for h, t in hog2tax.items()}

        LOG.debug("fill family and HOG tables")
        self.update_hog_and_fam_tabs(fam2hogs, hog2taxoff, hog2protoffs, hog2oma_hog)

        # add family and hog offsets
        LOG.debug("complete protein table")
        self.update_prot_tab(hog2protoffs)

        LOG.debug("compute LCA taxa")
        self.add_lcataxoff_col()

        # close and open in read mode
        h5file.close()

    ### functions to parse OMA database ###
    def select_and_strip_OMA_HOGs(self, h5file):
        def _process_oma_hog(
            tax2level,
            curr_oma_taxa,
            curr_oma_hog,
            curr_oma_roothog,
            fam,
            fam2hogs,
            hog2oma_hog,
            hog2tax,
            roottax,
            include_younger_fams
        ):
            """
			- decide whether an OMA HOG should be stored based on current root-HOG and HOG taxa
			- update root-HOG if necessary and keep track of HOG taxa
			"""

            def _is_descendant(hog1, hog2):
                """
			    True if hog1 descendant of hog2
			    """
                return hog2 in self.parse_hogs(hog1)

            def _store(
                fam, curr_oma_hog, curr_oma_roothog, fam2hogs, hog2oma_hog, hog2tax, tax
            ):
                """
			    - compute SPhog HOG id
			    - fill containers
			    """
                hog = b".".join(
                    [
                        str(fam).encode("ascii"),
                        *curr_oma_hog.split(b".")[len(curr_oma_roothog.split(b".")) :],
                    ]
                )

                fam2hogs[fam].add(hog)
                hog2oma_hog[hog] = curr_oma_hog
                hog2tax[hog] = tax

            # compute most ancestral taxon; when absent, flag it with 1000000
            tax_levels = list(map(lambda x: tax2level.get(x, 1000000), curr_oma_taxa))
            min_level = min(tax_levels)

            if min_level == 1000000:
                tax = None
            else:
                tax = curr_oma_taxa[tax_levels.index(min_level)]

            # store if descendant of current OMA root-HOG
            if _is_descendant(curr_oma_hog, curr_oma_roothog):

                _store(
                    fam,
                    curr_oma_hog,
                    curr_oma_roothog,
                    fam2hogs,
                    hog2oma_hog,
                    hog2tax,
                    tax,
                )

            # if tax equal or younger that root taxon: store, update fam and current OMA root-HOG
            elif tax:

                # include only family at root-taxon if include_younger_fams==False
                if (not include_younger_fams and tax == roottax) or include_younger_fams:

                    fam += 1
                    curr_oma_roothog = curr_oma_hog

                    # store after updating fam
                    _store(
                        fam,
                        curr_oma_hog,
                        curr_oma_roothog,
                        fam2hogs,
                        hog2oma_hog,
                        hog2tax,
                        tax,
                    )

            return fam, curr_oma_roothog

        def _process_oma_fam(
            fam_tab_sort, tax2level, fam, fam2hogs, hog2oma_hog, hog2tax, roottax, include_younger_fams
        ):
            """
			apply _process_oma_hog to one OMA family
			 - fam_tab_sort: slice of the HogLevel table for one family sorted by HOG ids
			 - tax2level: mapper between taxa and their distance from roottax
			"""
            # bookeepers for HOGs
            curr_oma_hog = fam_tab_sort[0]["ID"]
            curr_oma_roothog = None
            curr_oma_taxa = []

            for r in fam_tab_sort:
                oma_hog, oma_tax = r[1], r[2] #b"".join(r[2].split())

                # evaluation at new oma HOG
                if oma_hog != curr_oma_hog:

                    fam, curr_oma_roothog = _process_oma_hog(
                        tax2level,
                        curr_oma_taxa,
                        curr_oma_hog,
                        curr_oma_roothog,
                        fam,
                        fam2hogs,
                        hog2oma_hog,
                        hog2tax,
                        roottax,
                        include_younger_fams
                    )

                    # reset for new HOG
                    curr_oma_taxa = []
                    curr_oma_hog = oma_hog

                # track taxa of current oma HOG
                curr_oma_taxa.append(oma_tax)

            # end
            fam, curr_oma_roothog = _process_oma_hog(
                tax2level,
                curr_oma_taxa,
                curr_oma_hog,
                curr_oma_roothog,
                fam,
                fam2hogs,
                hog2oma_hog,
                hog2tax,
                roottax,
                include_younger_fams
            )

            return fam

        #
        tax2level = dict(zip(self._tax_tab[:]["ID"], self._tax_tab[:]["Level"]))
        hog_tab = h5file.root.HogLevel

        # load families in memory
        families = hog_tab.col("Fam")

        # containers
        fam2hogs = collections.defaultdict(set)
        hog2oma_hog = dict()
        hog2tax = dict()

        # bookeepers for families
        fam = 0
        curr_fam = families[0]
        i = 0
        j = 0

        for oma_fam in tqdm(families, disable=is_progress_disabled()):
            if oma_fam != curr_fam:

                # load fam table and sort by HOG ids
                fam_tab = hog_tab[i:j]
                fam_tab = np.sort(fam_tab, order="ID")

                # select and strip HOGs of one family
                fam = _process_oma_fam(
                    fam_tab,
                    tax2level,
                    fam,
                    fam2hogs,
                    hog2oma_hog,
                    hog2tax,
                    self.root_taxon.encode("ascii"),
                    self.include_younger_fams
                )

                # move pointer and update current family
                i = j
                curr_fam = oma_fam
            j += 1

        # end
        fam_tab = hog_tab[i:j]
        fam_tab = np.sort(fam_tab, order="ID")
        fam = _process_oma_fam(
            fam_tab,
            tax2level,
            fam,
            fam2hogs,
            hog2oma_hog,
            hog2tax,
            self.root_taxon.encode("ascii"),
            self.include_younger_fams
        )

        del hog_tab, families

        return fam2hogs, hog2oma_hog, hog2tax

    def select_and_filter_OMA_proteins(
        self, h5file, fam2hogs, hog2oma_hog, hog2tax, species, min_prot_nr
    ):
        """
        One small diff compared to DatabaseFromFasta is that proteins in protein table are not following the species in species table. This should be fine.
        """
        genome_tab = h5file.root.Genome[:]
        ent_tab = h5file.root.Protein.Entries
        oma_seq_buffer = h5file.root.Protein.SequenceBuffer

        # temporary mappers
        oma_hog2hog = dict(zip(hog2oma_hog.values(), hog2oma_hog.keys()))
        spe2speoff = dict(
            zip(sorted(species), range(len(species)))
        )  # this is to sort the species table

        LOG.debug(" - select proteins from selected HOGs")
        # remember the size of families and bookkeep family of each protein for filtering
        fam2prot_nr = collections.Counter()
        prot_fams = []

        # bookkeep HOG of each protein to later build the hog2protoffs mapper
        prot_hogs = []

        # bookkeep sequence offset of each protein to later build the local seq_buff
        oma_seq_offsets = []

        # store temporary rows for species and protein tables
        oma_species = [b""] * len(species)  # to keep species sorted
        prot_rows = []

        for r in tqdm(genome_tab, disable=is_progress_disabled()):

            sp = r['SciName']
            sp_code = r['UniProtSpeciesCode']
            
            # use species code if scientific name is lacking
            # for ~27 cases the uniprot id replaces the scientific name in OMA species tree
            if sp not in species:
                sp = sp_code

            # filter if species outside root-taxon
            if sp in species:
                entry_off = r["EntryOff"]
                entry_num = r["TotEntries"]

                spe_off = spe2speoff[sp]
                oma_species[spe_off] = sp

                # sub entry table for species
                sp_ent_tab = ent_tab[entry_off : entry_off + entry_num]

                for rr in sp_ent_tab:
                    oma_hog = rr["OmaHOG"]
                    # select protein if member of selected OMA HOG
                    if oma_hog in oma_hog2hog:
                        # update counter and track hogs and families
                        hog = oma_hog2hog[oma_hog]
                        fam = int(hog.split(b".")[0].decode("ascii"))
                        fam2prot_nr[fam] += 1
                        prot_hogs.append(hog)
                        prot_fams.append(fam)

                        # track sequence length and offset
                        oma_seq_off = rr["SeqBufferOffset"]
                        seq_len = rr["SeqBufferLength"]
                        oma_seq_offsets.append(oma_seq_off)

                        # store protein row
                        oma_id = '{}{:05d}'.format(sp_code.decode('ascii'), rr['EntryNr'] - entry_off)
                        prot_rows.append((oma_id.encode('ascii'), 0, seq_len, spe_off, 0, 0))

        LOG.debug(" - filter by family protein number")
        # now filter the mappers
        f_fam2hogs = collections.defaultdict(set)
        f_hog2oma_hog = dict()
        f_hog2tax = dict()

        for fam, hogs in fam2hogs.items():

            prot_nr = fam2prot_nr[fam]

            # filter by size
            if prot_nr >= min_prot_nr:

                f_fam2hogs[fam] = hogs
                f_hog2oma_hog.update(
                    dict(zip(hogs, list(map(lambda x: hog2oma_hog[x], hogs))))
                )
                f_hog2tax.update(dict(zip(hogs, list(map(lambda x: hog2tax[x], hogs)))))

        LOG.debug(" - filter proteins based on filtered families")
        # bookkeeping for later
        hog2protoffs = collections.defaultdict(set)

        # pointer to species in species table
        curr_spe_off = prot_rows[0][3]

        # pointer to protein in protein table
        prot_off = self._prot_tab.nrows
        curr_prot_off = prot_off

        # pointer to sequence in buffer
        seq_off = self._seq_buff.nrows

        # store rows for species and protein tables	and sequence buffer
        spe_rows = [()] * len(species)  # keep sorted
        f_prot_rows = []
        seq_buff = []

        # iter trough temporary protein rows
        i = 0
        for r in tqdm(prot_rows, disable=is_progress_disabled()):
            if prot_fams[i] in f_fam2hogs:
                spe_off = r[3]

                # change of species --> store current species
                if spe_off != curr_spe_off:

                    spe_rows[curr_spe_off] = (
                        oma_species[curr_spe_off],
                        curr_prot_off,
                        prot_off - curr_prot_off,
                        0,
                    )

                    curr_prot_off = prot_off
                    curr_spe_off = spe_off

                    # take advantage to dump sequence buffer
                    self._seq_buff.append(seq_buff)
                    self._seq_buff.flush()
                    seq_buff = []

                # sequence
                oma_seq_off = oma_seq_offsets[i]
                seq_len = r[2]
                seq = oma_seq_buffer[oma_seq_off : oma_seq_off + seq_len]
                seq_buff.extend(list(seq))

                # store protein row
                f_prot_rows.append((r[0], seq_off, seq_len, spe_off, 0, 0))

                # book keeping
                hog2protoffs[prot_hogs[i]].add(prot_off)

                # update offset of protein sequence in buffer and of protein raw in table
                seq_off += seq_len
                prot_off += 1
            i += 1

        # end
        spe_rows[curr_spe_off] = (
            oma_species[curr_spe_off],
            curr_prot_off,
            prot_off - curr_prot_off,
            0,
        )

        self._seq_buff.append(seq_buff)
        self._seq_buff.flush()

        # fill species and protein tables
        self._sp_tab.append(spe_rows)
        self._sp_tab.flush()
        self._prot_tab.append(f_prot_rows)
        self._prot_tab.flush()

        del genome_tab, ent_tab, oma_seq_buffer

        return f_fam2hogs, hog2protoffs, f_hog2tax, f_hog2oma_hog



import os
import sys
import numba
import scipy
import scipy.special
import tables
import numpy as np
import pandas as pd
from property_manager import lazy_property, cached_property

from .hierarchy import _children_hog

class ProbSearch(object):
    def __init__(self, fs):
        
        assert fs.db.mode == 'r', 'Database must be opened in read mode.'
        assert fs.ki.mode == 'r', 'Index must be opened in read mode.'
        
        # load fs, ki and db
        self.db = fs.db
        self.ki = fs.ki
        self.fs = fs
        
    @lazy_property
    def nr_kmer(self):
        return np.unique(self.ki._table_idx[:]).size -1

    def search(self, top_n=10):
        '''
        args
         - top_n: top n families on which compute the score and if cum_mode='max', also on which to cumulate counts.
         - cum_mode: how count are cumulated inside families. They can be summed or maxed at each node.
         - score: how counts are normalized: naive, querysize, theo_prob, kmerfreq_prob, etc.
        '''
        assert top_n <= len(self.db._fam_tab), 'top n is smaller than number of families'

        print('compute family scores')
        # filter top n families
        fam_ranked_n = self.fs._fam_ranked[:][:,:top_n]

        # cumulated queryHOG counts for the top n families
        queryHog_cum_counts = self.cumulate_counts_nfams_nqueries(
            fam_ranked_n, self.fs._queryHog_count[:], self.db._fam_tab[:], self.db._level_arr[:], self.db._hog_tab.col('ParentOff'), 
            len(self.fs._query_count), self.cumulate_counts_1fam, self._sum, self._max)

        # update queryFam counts to have maxed counts instead of summed counts (using root-HOG maxed counts)
        fam_rh_offsets = self.db._fam_tab.col('HOGoff')[fam_ranked_n]
        queryFam_cum_counts = queryHog_cum_counts[np.arange(fam_rh_offsets.shape[0])[:,None], fam_rh_offsets]

        # cumulate HOG counts
        hog_cum_counts = self.cumulate_counts_nfams(
            self.ki._hog_count[:], self.db._fam_tab[:], self.db._level_arr[:], self.db._hog_tab.col('ParentOff'), self.cumulate_counts_1fam, self._sum, self._max)            

        # and get family cumulated counts
        fam_cum_counts = hog_cum_counts[self.db._fam_tab.col('HOGoff')]

        # normalize family cum counts
        queryFam_scores = self.compute_family_kmerfreq_probs(
            fam_ranked_n, self.fs._query_count[:], self.fs._query_occur[:], queryFam_cum_counts, fam_cum_counts, self.nr_kmer)

        # ranked families
        fam_reranked_1, queryFam_scores = self.reranked_families_and_scores(queryFam_scores, fam_ranked_n, True)

        print('compute subfamily scores')
        ### Subfamilies
        queryHog_scores, queryHog_bestpaths = self.compute_subfamily_kmerfreq_probs(
            fam_reranked_1, self.fs._query_count[:], self.fs._query_occur[:], self.db._fam_tab[:], queryFam_scores, self.db._hog_tab[:], 
            self.db._chog_arr[:], queryHog_cum_counts, hog_cum_counts, self.fs._queryHog_count[:], self.fs._queryHog_occur[:], self.nr_kmer)

        #print("store results")
        # get the results -- do something a bit stupid here to begin with...

        # - queryHog_score is matrix of scores to output
        # - bestpath_mask contains a mask of this?

        qseqid = self.fs._query_id[:].flatten()  # these will come from FASTA file
        family = fam_reranked_1[:, 0]                # family offset id
        score = queryFam_scores[:, 0]
        f = np.isnan(score)
        family = list(map(lambda x: self.db._hog_tab[x]['OmaID'].decode('ascii'),  # TODO: update this to be HOGID from OXML
                          map(lambda x: self.db._fam_tab[x]['HOGoff'],
                              family[~f])))


        # change this so that we don't load all into memory...
        df = pd.DataFrame({'qseqid': qseqid[~f],  # map these to the id?
                           'family': family,  # need to map these to the final ID
                           'score': score[~f]})
        #df[['qseqid', 'family', 'score']].to_csv('out.tsv', sep='\t', index=False, float_format='%.2f')

        # sub-family results
        def get_res():
            for x in np.argwhere(queryHog_bestpaths):
                ii = qseqid[x[0]]
                h = self.db._hog_tab[x[1]]['OmaID']
                s = queryHog_scores[x[0],x[1]]
                if not np.isnan(s):
                    yield {'qseqid': ii, 'hog': h, 'score': s}

        prots = get_descendant_prots(get_descendant_hogs(0, hog_tab, chog_buff), hog_tab, cprot_buff)

        df1 = pd.DataFrame.from_records(get_res())

        return (df, df1)

    def compute_family_kmerfreq_probs(self, fam_ranked_n, query_counts, query_occurs, queryFam_cum_count, fam_cum_counts, nr_kmer):

        queryFam_score = np.zeros(fam_ranked_n.shape, np.float64)

        for q in range(fam_ranked_n.shape[0]):

            # number of unique k-mer in the query
            query_size = query_counts[q]

            # sum of k-mer family occurence of the query
            query_occur = query_occurs[q]
            
            for i in range(fam_ranked_n.shape[1]):
                
                # family == root-HOG
                qf_ccount = queryFam_cum_count[q, i]
                f_ccount = fam_cum_counts[fam_ranked_n[q, i]]
                
                queryFam_score[q, i] = self.compute_prob_score(
                    query_size, qf_ccount, self._bernoulli_true_kmerfreq, 
                    query_occur=query_occur, query_count=query_size, nr_kmer=nr_kmer, hog_count=f_ccount)
        
        return queryFam_score

    def compute_prob_score(self, query_size, qh_ccount, bernoulli_fun, **kwargs):

        # compute the binomial bernoulli probability (P(draw a k-mer in the HOG))
        bernoulli = bernoulli_fun(**kwargs)

        # probabilities for tail x values
        tail_size = query_size - qh_ccount 
        tail_log_probs = self.poisson_log_pmf(np.arange(qh_ccount, query_size + 1), np.full(tail_size + 1, bernoulli * query_size))

        # sum of these tail probabilities
        return scipy.special.logsumexp(tail_log_probs)
    
    @staticmethod
    def poisson_log_pmf(k, lda):
        return k*scipy.log(lda) - lda - scipy.special.gammaln(k + 1)

    @staticmethod
    def reranked_families_and_scores(queryFam_scores, fam_ranked_n, prob=True):
        if prob:
            idx = queryFam_scores.argsort()
        else:
            idx = (-queryFam_scores).argsort()
        return fam_ranked_n[np.arange(fam_ranked_n.shape[0])[:,None], idx][:,:1], queryFam_scores[np.arange(fam_ranked_n.shape[0])[:,None], idx][:,:1]

    @staticmethod
    def _bernoulli_true_kmerfreq(query_occur, query_count, nr_kmer, hog_count):
        '''
        true k-mer frequencies are used for probability to get a k-mer at one location
        Bernoulli is the joint probability of not having the k-mer in the family
        '''
        kmer_prob = query_occur / query_count / nr_kmer

        return 1 - (1 - kmer_prob) ** hog_count

    ### subfamily ######################################################################################################################
    def compute_subfamily_kmerfreq_probs(
        self, fam_reranked_1, query_counts, query_occurs, fam_tab, queryFam_scores, hog_tab, chog_buff, queryHog_cum_counts, hog_cum_counts, queryHog_counts, queryHog_occurs, nr_kmer):
        
        def _top_down_best_path(
            self, hog_off, hog_tab, chog_buff, queryHog_cum_counts, query_off, hog_cum_counts, query_occur, query_size, nr_kmer, queryHog_counts, queryHog_occurs,
            queryHog_score, queryHog_bestpath):
        
            if hog_off and hog_tab[hog_off]['ChildrenHOGoff'] != -1:

                # get HOG children
                children = _children_hog(hog_off, hog_tab, chog_buff)

                # intiate scores; either raw cumulative counts or probabilities
                child_scores = np.zeros(children.shape)

                # compute scores
                for i, h in enumerate(children):

                    # query-HOG cumulative k-mer counts
                    qh_ccount = queryHog_cum_counts[query_off, h]

                    if qh_ccount > 0:
                        h_ccount = hog_cum_counts[h]
                        
                        child_scores[i] = self.compute_prob_score(
                            query_size, qh_ccount, self._bernoulli_true_kmerfreq, 
                            query_occur=query_occur, query_count=query_size, nr_kmer=nr_kmer, hog_count=h_ccount)

                    elif qh_ccount == 0:
                        child_scores[i] = 0.0  # because log(1) = 0

                # find best child
                cand_offsets = np.where((child_scores < 0) & (child_scores==np.min(child_scores)))[0]

                # deals with ties
                if cand_offsets.size == 1:
                    best_child = children[cand_offsets][0] 

                    # store best path and score --> latter use buff for that
                    queryHog_bestpath[query_off, best_child] = True
                    queryHog_score[query_off, best_child] = child_scores[cand_offsets][0]

                else:
                    best_child = None       

                # remove queryHOG count from query size (not the cumulated one!)
                query_size = query_size - queryHog_counts[query_off, best_child]

                # remove queryHOG k-mer occurences (sum of k-mer occurences for the intersecting k-mers) from query occur
                query_occur = query_occur - queryHog_occurs[query_off, best_child]
                
                _top_down_best_path(
                    self, best_child, hog_tab, chog_buff, queryHog_cum_counts, query_off, hog_cum_counts, query_occur, query_size, nr_kmer, queryHog_counts, queryHog_occurs, 
                    queryHog_score, queryHog_bestpath)
        
        queryHog_scores = np.zeros(queryHog_cum_counts.shape, dtype=np.float64)
        queryHog_bestpaths = np.full(queryHog_cum_counts.shape, False)
        
        for q in range(fam_reranked_1.shape[0]):

            # number of unique k-mer in the query
            query_size = query_counts[q]

            # sum of k-mer family occurence of the query
            query_occur = query_occurs[q]

            rh = fam_tab[fam_reranked_1[q, 0]]['HOGoff']

            # add root-HOG score and best path
            queryHog_scores[q, rh] = queryFam_scores[q, 0]
            queryHog_bestpaths[q, rh] = True

            _top_down_best_path(
                self, rh, hog_tab, chog_buff, queryHog_cum_counts, q, hog_cum_counts, query_occur, query_size, nr_kmer, queryHog_counts, queryHog_occurs, 
                queryHog_scores, queryHog_bestpaths)
            
        return queryHog_scores, queryHog_bestpaths

    ### cumulate counts #####################################################################################################################
    @staticmethod
    @numba.njit
    def _max(x, y):
        return max(x, y)

    @staticmethod
    @numba.njit
    def _sum(x, y):
            return x + y 

    @staticmethod
    @numba.njit
    def cumulate_counts_1fam(hog_cum_counts, fam_level_offsets, hog2parent, cum_fun, prop_fun):
        
        current_best_child_count = np.zeros(hog_cum_counts.shape, dtype=np.uint64)
        
        # iterate over level offsets backward
        for i in range(fam_level_offsets.size - 2):
            x = fam_level_offsets[-i - 3: -i - 1]

            # when reaching level, sum all hog counts with their best child count
            hog_cum_counts[x[0]:x[1]] = cum_fun(hog_cum_counts[x[0]:x[1]], current_best_child_count[x[0]:x[1]])   

            # update current_best_child_count of the parents of the current hogs
            for i in range(x[0], x[1]):
                parent_off = hog2parent[i]
                
                # only if parent exists
                if parent_off != -1:
                    c = current_best_child_count[hog2parent[i]]
                    current_best_child_count[hog2parent[i]] = prop_fun(c, hog_cum_counts[i])
    
    @staticmethod
    @numba.njit(parallel=True, nogil=True)
    def cumulate_counts_nfams(hog_counts, fam_tab, level_arr, hog2parent, main_fun, cum_fun, prop_fun):

        hog_cum_counts = hog_counts.copy()

        for fam_off in numba.prange(fam_tab.size):
            entry = fam_tab[fam_off]
            level_off = entry['LevelOff']
            level_num = entry['LevelNum']
            fam_level_offsets = level_arr[level_off:np.int64(level_off + level_num + 2)]

            main_fun(hog_cum_counts, fam_level_offsets, hog2parent, cum_fun, prop_fun)

        return hog_cum_counts

    @staticmethod
    @numba.njit(parallel=True, nogil=True)
    def cumulate_counts_nfams_nqueries(fam_results, hog_counts, fam_tab, level_arr, hog2parent, q_num, main_fun, cum_fun, prop_fun):
        
        hog_cum_counts = hog_counts.copy()
        
        # iterate queries
        for q in numba.prange(q_num):
            
            # iterate families
            for fam_off in fam_results[q]:
                
                entry = fam_tab[fam_off]
                level_off = entry['LevelOff']
                level_num = entry['LevelNum']
                fam_level_offsets = level_arr[level_off:np.int64(level_off + level_num + 2)]
                main_fun(hog_cum_counts[q], fam_level_offsets, hog2parent, cum_fun, prop_fun)
            
        return hog_cum_counts



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from property_manager import lazy_property
import numba
import numpy as np
import os
import pandas as pd
import sys

from ._utils import LOG
from .hierarchy import get_descendant_hogs, get_descendant_prots


class Search(object):
    def __init__(self, fs, include_extant_genes=False):
        assert fs.db.mode == "r", "Database must be opened in read mode."

        # load fs, ki and db
        self.db = fs.db
        self.ki = fs.ki
        self.fs = fs

        self.include_extant_genes = include_extant_genes

    #@lazy_property
    #def hog_cum_count(self):
    #    return self.cumulate_counts_nfams(
    #        self.ki._hog_count[:],
    #        self.db._fam_tab[:],
    #        self.db._level_arr[:],
    #        self.db._hog_tab.col("ParentOff"),
    #        self.cumulate_counts_1fam,
    #        self._sum,
    #        self._max,
    #    )

    #@lazy_property
    #def total_occur(self):
    #    return self.ki._idx_arr[:].size

    def search(
        self,
        norm_fam_fun,
        top_n,
        prop_fun,
        norm_hog_fun,
        threshold,
        norm_fam_args=[],
        norm_hog_args=[]):
        """
        args
         - norm_fam_fun: choice of function to normalize k-mer count of families
         - top_n: number of families to proceed with cumulation of HOG k-mer counts
         - prop_fun: function to decide how counts are cumulated (_max, _sum)
         - norm_hog_fun: choice of function to normalize k-mer count of HOGs

        additional args can be passed to norm_fam_fun and norm_hog_fun
        """
        top_n = min(top_n, len(self.db._fam_tab))
        #assert (
        #    top_n <= self.db._fam_tab.nrows
        #), "top n must be smaller than total number of families"

        # normalize at family level
        LOG.debug("compute family scores")
        queryFam_score = norm_fam_fun(*norm_fam_args)

        # resort families after normalization and keep top n
        fam_ranked = self.resort_ranked_fam(self.fs._fam_ranked[:], queryFam_score)[
            :, :top_n
        ]

        LOG.debug("propagate HOG counts bottom-up")
        queryHog_cum_count = self.cumulate_counts_nfams_nqueries(
            fam_ranked,
            self.fs._queryHog_count[:],
            self.db._fam_tab[:],
            self.db._level_arr[:],
            self.db._hog_tab.col("ParentOff"),
            len(self.fs._query_count),
            self.cumulate_counts_1fam,
            self._sum,
            self._max,
        )

        LOG.debug("compute HOG scores")
        queryHog_score, queryHog_bestpath = norm_hog_fun(
            fam_ranked, queryHog_cum_count, *norm_hog_args
        )

        LOG.debug("form results table")
        # get the results -- do something a bit stupid here to begin with...

        # - queryHog_score is matrix of scores to output
        # - bestpath_mask contains a mask of this? 
        def get_prot_ids(h):
            desc_hogs = list(get_descendant_hogs(h, self.db._hog_tab, self.db._chog_arr)) + [h]
            prot_ii = get_descendant_prots(desc_hogs, self.db._hog_tab, self.db._cprot_arr)
            return self.db._prot_tab.read_coordinates(prot_ii, 'ID')
        
        qseqid = self.fs._query_id[:].flatten()  # these will come from FASTA file
        family = fam_ranked[:, 0]                # family offset id
        family_score = queryFam_score[np.arange(len(qseqid)), family] 

        # generate the dataframe
        def generate_results(threshold):
            for i in np.argwhere(~np.isnan(family_score)).flatten():
                # find best scoring subfamily
                best_j = None
                best_s = np.inf
                for j in np.argwhere(queryHog_bestpath[i]).flatten():
                    s = queryHog_score[i,j]
                    if s < best_s and s >= threshold:
                        best_j = j
                        best_s = s
                if best_j is not None:
                    z = {'qseqid': qseqid[i],
                         'family': self.db._hog_tab[self.db._fam_tab[family[i]]['HOGoff']]['OmaID'].decode('ascii'),
                         'family-score': family_score[i],
                         'subfamily': self.db._hog_tab[best_j]['OmaID'].decode('ascii'),
                         'subfamily-score': best_s}
                    if self.include_extant_genes:
                        z['subfamily-geneset'] = ','.join(map(lambda x: x.decode('ascii'), get_prot_ids(best_j)))
                    yield z

        h = ['qseqid', 'family', 'family-score', 'subfamily', 'subfamily-score']
        if self.include_extant_genes:
            h.append('subfamily-geneset')

        return pd.DataFrame(generate_results(threshold))[h]

    ### normalize family ####################################################################################################################
    def norm_fam_query_size(self):
        """
        simple division by the length of query
        """
        # no need to broadcast because query_count is a vertical vector
        # query_count_bc = np.full((self.fs._queryFam_count[:].shape[1], self.fs._queryFam_count[:].shape[0]), self.fs._query_count[:]).T

        return self.fs._queryFam_count[:] / self.fs._query_count[:]

    ### sort families #######################################################################################################################
    @staticmethod
    def resort_ranked_fam(fam_ranked, queryFam_score, top_n=None):

        # ranked scores following fam_ranked until top_n
        ranked_score = queryFam_score[
            np.arange(fam_ranked.shape[0])[:, None], fam_ranked[:, :top_n]
        ]

        # get indices of the reverse sorted ranked scores
        idx = (-ranked_score).argsort()

        # back to the fam offsets by indexing fam_ranked with our indices
        return fam_ranked[np.arange(fam_ranked.shape[0])[:, None], idx]

    ### cumulate counts #####################################################################################################################
    @staticmethod
    @numba.njit
    def _max(x, y):
        return max(x, y)

    @staticmethod
    @numba.njit
    def _sum(x, y):
        return x + y

    @staticmethod
    @numba.njit
    def cumulate_counts_1fam(
        hog_cum_counts, fam_level_offsets, hog2parent, cum_fun, prop_fun
    ):

        current_best_child_count = np.zeros(hog_cum_counts.shape, dtype=np.uint16)

        # iterate over level offsets backward
        for i in range(fam_level_offsets.size - 2):
            x = fam_level_offsets[-i - 3 : -i - 1]

            # when reaching level, sum all hog counts with their best child count
            hog_cum_counts[x[0] : x[1]] = cum_fun(
                hog_cum_counts[x[0] : x[1]], current_best_child_count[x[0] : x[1]]
            )

            # update current_best_child_count of the parents of the current hogs
            for i in range(x[0], x[1]):
                parent_off = hog2parent[i]

                # only if parent exists
                if parent_off != -1:
                    c = current_best_child_count[hog2parent[i]]
                    current_best_child_count[hog2parent[i]] = prop_fun(
                        c, hog_cum_counts[i]
                    )

    @staticmethod
    @numba.njit(parallel=True, nogil=True)
    def cumulate_counts_nfams(
        hog_counts, fam_tab, level_arr, hog2parent, main_fun, cum_fun, prop_fun
    ):

        hog_cum_counts = hog_counts.copy()

        for fam_off in numba.prange(fam_tab.size):
            entry = fam_tab[fam_off]
            level_off = entry["LevelOff"]
            level_num = entry["LevelNum"]
            fam_level_offsets = level_arr[
                level_off : np.int64(level_off + level_num + 2)
            ]

            main_fun(hog_cum_counts, fam_level_offsets, hog2parent, cum_fun, prop_fun)

        return hog_cum_counts

    @staticmethod
    @numba.njit(parallel=True, nogil=True)
    def cumulate_counts_nfams_nqueries(
        fam_results,
        hog_counts,
        fam_tab,
        level_arr,
        hog2parent,
        q_num,
        main_fun,
        cum_fun,
        prop_fun,
    ):

        hog_cum_counts = hog_counts.copy()  # probably not necessary

        # iterate queries
        for q in numba.prange(q_num):

            # iterate families
            for fam_off in fam_results[q]:

                entry = fam_tab[fam_off]
                level_off = entry["LevelOff"]
                level_num = entry["LevelNum"]
                fam_level_offsets = level_arr[
                    level_off : np.int64(level_off + level_num + 2)
                ]
                main_fun(
                    hog_cum_counts[q], fam_level_offsets, hog2parent, cum_fun, prop_fun
                )

        return hog_cum_counts

    ### normalize hogs ######################################################################################################################
    def norm_hog_query_size(self, fam_ranked, queryHog_cum_count):
        """
        same but normalize first the queryHog_cum_count
        could be slightly more clever and remove part conserved in parent 
        """
        # query_count_bc = np.full((queryHog_cum_count.shape[1], queryHog_cum_count.shape[0]), self.fs._query_count[:]).T

        queryHog_score = queryHog_cum_count / self.fs._query_count[:]

        queryHog_bestpath = self.compute_bestpath_nqueries_nfams(
            len(self.fs._query_count),
            fam_ranked,
            queryHog_score,
            self.db._fam_tab[:],
            self.db._level_arr[:],
            self.db._hog_tab.col("ParentOff"),
        )

        return queryHog_score, queryHog_bestpath

    @staticmethod
    def store_bestpath(
        hog_offsets, parent_offsets, queryHog_bestpath, queryHog_score, pv=False
    ):

        # keep HOGs descending from the best path
        cands = hog_offsets[queryHog_bestpath[parent_offsets]]

        # get the score of these candidates
        cands_scores = queryHog_score[cands]

        # find the candidate HOGs offsets with the higher count (>0) at this level
        if cands_scores.size > 0:

            # need smallest
            if pv:
                cands_offsets = np.where(cands_scores == np.min(cands_scores))[0]
            # need > 0 and max
            else:
                cands_offsets = np.where(
                    (cands_scores > 0) & (cands_scores == np.max(cands_scores))
                )[0]

            # if a single candidate, update the best path. Else, stop because of tie
            if cands_offsets.size == 1:
                queryHog_bestpath[cands[cands_offsets]] = True

    def compute_bestpath_nqueries_nfams(
        self, q_num, fam_ranked, queryHog_cum_count, fam_tab, level_arr, hog2parent
    ):
        """
        this one is to compute the best path w/o computing the score simultaneously
        """
        queryHog_bestpath = np.full(queryHog_cum_count.shape, False)

        # loop through queries and families
        for q in range(q_num):
            for f in fam_ranked[q]:

                entry = fam_tab[f]
                level_off = entry["LevelOff"]
                level_num = entry["LevelNum"]
                fam_level_offsets = level_arr[
                    level_off : np.int64(level_off + level_num + 2)
                ]

                # use the cumulative counts to define the best path
                rh = fam_level_offsets[0]
                qh_score = queryHog_cum_count[q][rh]
                if qh_score > 0:
                    queryHog_bestpath[q][rh] = True

                # loop through hog levels
                for i in range(1, fam_level_offsets.size - 2):
                    x = fam_level_offsets[i : i + 2]
                    hog_offsets = np.arange(x[0], x[1])

                    # grab parents
                    parent_offsets = hog2parent[hog_offsets]

                    # store best path
                    self.store_bestpath(
                        hog_offsets,
                        parent_offsets,
                        queryHog_bestpath[q],
                        queryHog_cum_count[q],
                    )

        return queryHog_bestpath



'''
    OMAmer - tree-driven and alignment-free protein assignment to sub-families

    (C) 2019-2020 Victor Rossier <victor.rossier@unil.ch> and
                  Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>

    This file is part of OMAmer.

    OMAmer is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OMAmer is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with OMAmer. If not, see <http://www.gnu.org/licenses/>.
'''
from datetime import date

__packagename__ = "omamer"
__version__ = "0.1.2"
__copyright__ = "(C) 2019-{:d} Victor Rossier <victor.rossier@unil.ch> and Alex Warwick Vesztrocy <alex@warwickvesztrocy.co.uk>".format(
    date.today().year
)

from .hierarchy import *
from .alphabets import *
from .database import *
from .index import *
from .flat_search import *
from .search import *
from .prob_search import *
from .merge_search import *
from .validation import *